# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.bebob.bebob_unit import BebobUnit

from hinawa_utils.bebob.maudio_protocol_normal import MaudioProtocolNormal
from hinawa_utils.bebob.maudio_protocol_fw410 import MaudioProtocolFw410
from hinawa_utils.bebob.maudio_protocol_special import MaudioProtocolSpecial

__all__ = ['MaudioUnit']


class MaudioUnit(BebobUnit):
    __SUPPORTED_MODELS = {
        # (VendorID, ModelID): Protocol
        (0x000d6c, 0x00000a): MaudioProtocolNormal,     # Ozonic
        (0x000d6c, 0x010062): MaudioProtocolNormal,     # Firewire Solo
        (0x000d6c, 0x010060): MaudioProtocolNormal,     # Firewire Audiophile
        (0x000d6c, 0x010081): MaudioProtocolNormal,     # NRV10
        (0x000d6c, 0x0100a1): MaudioProtocolNormal,     # Profire Lightbridge
        (0x0007f5, 0x010046): MaudioProtocolFw410,      # Firewire 410
        (0x000d6c, 0x010071): MaudioProtocolSpecial,    # Firewire 1814
        (0x000d6c, 0x010091): MaudioProtocolSpecial,    # ProjectMix I/O
    }

    def __init__(self, path):
        super().__init__(path)

        key = (self.vendor_id, self.model_id)
        if key not in self.__SUPPORTED_MODELS:
            raise OSError('Not supported.')
        self.protocol = self.__SUPPORTED_MODELS[key](self, False)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import unpack
from enum import Enum, auto

from hinawa_utils.ieee1212.config_rom_lexer import EntryType

__all__ = ['Ieee1212RootDirectoryParser']


class DirectoryContext(Enum):
    VENDOR = auto()
    SPECIFIER = auto()
    BUS_DEPENDENT = auto()
    KEYWORD = auto()


class DescriptorType(Enum):
    TEXTUAL = 0x00
    ICON = 0x01

    @classmethod
    def check_value(cls, value):
        return value in (item.value for item in cls)

#
# Table 16 - Key definitions
#


class KeyType(Enum):
    ROOT = 0x00  # For my convenience.
    DESCRIPTOR = 0x01
    BUS_DEPENDENT_INFO = 0x02
    VENDOR = 0x03
    HARDWARE_VERSION = 0x04
    MODULE = 0x07
    NODE_CAPABILITIES = 0x0c
    EUI_64 = 0x0d
    UNIT = 0x11
    SPECIFIER_ID = 0x12
    VERSION = 0x13
    DEPENDENT_INFO = 0x14
    UNIT_LOCATION = 0x15
    MODEL = 0x17
    INSTANCE = 0x18
    KEYWORD = 0x19
    FEATURE = 0x1a
    MODIFIABLE_DESCRIPTOR = 0x1f
    DIRECTORY_ID = 0x20

    @classmethod
    def check_value(cls, value):
        return value in (item.value for item in cls)


class Ieee1212RootDirectoryParser():
    #
    # Table 16 - Key definitions
    #
    _COMMON_KEYS = {
        # key:  (name, available types of parser)
        KeyType.DESCRIPTOR:             (EntryType.LEAF,
                                         EntryType.DIRECTORY),
        # 0x02 is for root directory only.
        KeyType.VENDOR:                 (EntryType.IMMEDIATE,
                                         EntryType.LEAF,
                                         EntryType.DIRECTORY),
        KeyType.HARDWARE_VERSION:       (EntryType.IMMEDIATE, ),
        # 0x05-06: reserved.
        # 0x07 is for root directory only.
        # 0x08-0b: reserved.
        # 0x07 is for root directory only.
        KeyType.EUI_64:                 (EntryType.LEAF, ),
        # 0x0e-10: reserved.
        # 0x11 is for root and instance directory only.
        KeyType.SPECIFIER_ID:           (EntryType.IMMEDIATE, ),
        KeyType.VERSION:                (EntryType.IMMEDIATE, ),
        KeyType.DEPENDENT_INFO:         (EntryType.IMMEDIATE,
                                         EntryType.CSR_OFFSET,
                                         EntryType.LEAF,
                                         EntryType.DIRECTORY, ),
        KeyType.UNIT_LOCATION:          (EntryType.LEAF, ),
        KeyType.MODEL:                  (EntryType.IMMEDIATE, ),
        # 0x18 and 0x19 are for root and instance directory only.
        # 0x21-2f: reserved.
        # 0x11 is for instance and unit directory only.
        # NOTE: keys for extended data are not supported.
        KeyType.MODIFIABLE_DESCRIPTOR: (EntryType.LEAF, ),
        KeyType.DIRECTORY_ID:          (EntryType.IMMEDIATE, ),
        # 0x30-37: by bus standard.
    }

    def __init__(self):
        self._bus_dep_handles = {}
        self._spec_dep_handles = {}
        self._vendor_dep_handles = {}
        self._keyword_dep_handles = {}

    def add_bus_dep_handle(self, name, handle):
        if name not in self._bus_dep_handles:
            self._bus_dep_handles[name] = []
        self._bus_dep_handles[name].append(handle)

    def add_spec_dep_handle(self, spec_id, version, handle):
        specifier = (spec_id, version)
        if specifier not in self._spec_dep_handles:
            self._spec_dep_handles[specifier] = []
        self._spec_dep_handles[specifier].append(handle)

    def add_vendor_dep_handle(self, vendor_id, handle):
        if vendor_id not in self._vendor_dep_handles:
            self._vendor_dep_handles[vendor_id] = []
        self._vendor_dep_handles[vendor_id].append(handle)

    def add_keyword_dep_handle(self, keyword, handle):
        if keyword not in self._keyword_dep_handles:
            self._keyword_dep_handles[keyword] = []
        self._keyword_dep_handles[keyword].append(handle)

    def __parse_immediate(self, key_type, ctx, value):
        #
        # 7.7.7 Node_Capabilities entry
        #
        if key_type == KeyType.NODE_CAPABILITIES:
            if self._bus_name in self._bus_dep_handles:
                for handle in self._bus_dep_handles[self._bus_name]:
                    data = handle(KeyType.NODE_CAPABILITIES.name,
                                  EntryType.IMMEDIATE.name, value)
                    if data:
                        value = data
                        break
        return value
    #
    # 7.5.4.1 Textual descriptors
    #

    def _parse_textual_descriptor(self, data):
        WIDTH_DEFINITIONS = {
            0x00:   'Fixed one-byte characters',
            0x01:   'Fixed two-byte characters',
            0x02:   'Fixed four-byte characters',
            # 0x03-07: reserved
            0x08:   'Not to be used',
            0x09:   'Variable width characters up to a two byte maximum',
            0x0a:   'Variable width characters up to a four byte maximum',
            # 0x0b-0f: reserved
        }

        width = data[0] >> 4
        character_set = ((data[0] & 0x0f) << 8) | data[1]
        language = unpack('>H', data[2:4])[0]

        if width not in WIDTH_DEFINITIONS or width != 0x00:
            raise OSError('Width {0} is not supported.'.format(width))

        # character_set == IANA MIBenum
        if character_set != 0:
            return OSError('Character set {0} is not supported.'.format(
                character_set))

        # At present, 'US-ASCII' is supported only.
        if language & 0x8000 or language > 0:
            raise OSError('Language {0} is not supported.'.format(language))

        content = data[4:].decode('US-ASCII') + '\0'
        return content[:content.find('\0')]

    #
    # 7.5.4.2 Icon descriptor
    #
    def _parse_icon_descriptor(self, data):
        raise OSError('Icon descriptor is not supported.')

    #
    # 7.5.4 Descriptors
    #
    def _parse_descriptor_leaf(self, data):
        TYPE_PARSERS = {
            DescriptorType.TEXTUAL: self._parse_textual_descriptor,
            DescriptorType.ICON:    self._parse_icon_descriptor,
        }

        descriptor_type = data[0]
        specifier_id = (data[1] << 16) | unpack('>H', data[2:4])[0]
        data = data[4:]

        if specifier_id == 0x00:
            if not DescriptorType.check_value(descriptor_type):
                raise OSError('Descritpor type {0} is not supported.'.format(
                    descriptor_type))
            type_id = DescriptorType(descriptor_type)
            return TYPE_PARSERS[type_id](data)

        if specifier_id in self._vendor_dep_handles:
            for handle in self._vendor_dep_handles[specifier_id]:
                elem = handle(KeyType.DESCRIPTOR.value, EntryType.LEAF.name,
                              data)
                if elem:
                    return elem

        return None

    def _parse_bus_dependent_info_leaf(self, data):
        # See explanation of Table 9 – Leaf format specifiers.
        if self._bus_name in self._bus_dep_handles:
            for handle in self._bus_dep_handles[self._bus_name]:
                elem = handle(KeyType.BUS_DEPENDENT_INFO.value,
                              EntryType.LEAF.name, data)
                if elem:
                    return elem
        return None

    def _parse_vendor_leaf(self, data):
        # TODO: handle VENDOR/SPECIFIER_ID in parent directory.
        if self._vendor_id in self._vendor_dep_handles:
            for handle in self._vendor_dep_handles[self._vendor_id]:
                elem = handle(KeyType.VENDOR.value, EntryType.IMMEDIATE.name,
                              data)
                if elem:
                    return elem
        return None

    #
    # 7.7.5 Module_Primary_EUI_64
    # 7.7.8 EUI_64 entry
    #
    def _parse_eui_64_leaf(self, data):
        info = (unpack('>I', data[0:4])[0] << 32) | unpack('>I', data[4:8])[0]
        return info

    def _parse_dependent_info_leaf(self, data):
        # TODO: handle VENDOR/SPECIFIER_ID in parent directory.
        if self._vendor_id in self._vendor_dep_handles:
            for handle in self._vendor_dep_handles[self._vendor_id]:
                elem = handle(KeyType.VENDOR.value, EntryType.IMMEDIATE.name,
                              data)
                if elem:
                    return elem
        return None

    #
    # 7.7.13 Unit_Location entry
    #
    def _parse_unit_location_leaf(self, data):
        info = {}

        info['base-address'] = \
            (unpack('>I', data[0:4])[0] << 32) | unpack('>I', data[4:8])[0]
        data = data[4:]
        info['upper-bound'] = \
            (unpack('>I', data[0:4])[0] << 32) | unpack('>I', data[4:8])[0]
        return info

    #
    # 7.6.5 Keyword leaves
    #
    def _parse_keyword_leaf(self, data):
        return data.rstrip('\0').decode('US-ASCII')

    def _parse_modifiable_desc_leaf(self, data):
        info = {}
        info['max_descriptor_size'] = unpack('>I', data[0:2])[0] * 4
        info['descriptor_address'] = \
            (unpack('>H', data[2:4])[0] << 32) | unpack('>I', data[4:8])[0]
        return info

    def _parse_leaf(self, key_type, ctx, data):
        LEAF_PARSERS = {
            KeyType.DESCRIPTOR:         self._parse_descriptor_leaf,
            KeyType.BUS_DEPENDENT_INFO: self._parse_bus_dependent_info_leaf,
            KeyType.VENDOR:             self._parse_vendor_leaf,
            KeyType.MODULE:             self._parse_eui_64_leaf,
            KeyType.EUI_64:             self._parse_eui_64_leaf,
            KeyType.DEPENDENT_INFO:     self._parse_dependent_info_leaf,
            KeyType.UNIT_LOCATION:      self._parse_unit_location_leaf,
            KeyType.KEYWORD:            self._parse_keyword_leaf,
            KeyType.MODIFIABLE_DESCRIPTOR: self._parse_modifiable_desc_leaf,
        }
        if key_type not in LEAF_PARSERS or not LEAF_PARSERS[key_type]:
            raise OSError('Key {0} is not supported.'.format(key_type))
        return LEAF_PARSERS[key_type](data)

    def _merge_common_keys(self, defined_keys):
        keys = self._COMMON_KEYS
        if defined_keys:
            keys = dict(keys)
            keys.update(defined_keys)
        return keys

    #
    # 7.5.4 Descriptors
    #
    def _parse_descriptor_directory(self, ctx, key_type, entries):
        # See annotation of Table 16 – Key definitions.
        DEFINED_KEYS = {
            KeyType.DESCRIPTOR:             (EntryType.LEAF,
                                             EntryType.DIRECTORY, ),
            KeyType.MODIFIABLE_DESCRIPTOR:  (EntryType.LEAF, ),
        }

        return self._parse_directory_entries(key_type, ctx, entries,
                                             DEFINED_KEYS)

    #
    # 7.7.1 Bus_Dependent_Info entry
    #
    def _parse_bus_dependent_directory(self, ctx, key_type, entries):
        # See explanation of Table 8 – Key ID allocations
        DEFINED_KEYS = {
            KeyType.BUS_DEPENDENT_INFO: (EntryType.IMMEDIATE,
                                         EntryType.CSR_OFFSET,
                                         EntryType.LEAF, ),
        }

        ctx = (DirectoryContext.BUS_DEPENDENT, self._NAME)

        return self._parse_directory_entries(key_type, ctx, entries,
                                             DEFINED_KEYS)

    #
    # 7.7.3 Vendor_Info entry
    #
    def _parse_vendor_directory(self, ctx, key_type, entries):
        # See explanation of Table 8 – Key ID allocations.
        for entry in entries:
            if entry[0] == (KeyType.SPECIFIER_ID.value, EntryType.IMMEDIATE):
                vendor_id = entry[1]
                break
        else:
            for entry in entries:
                if entry[0] == (KeyType.VENDOR.value, EntryType.IMMEDIATE):
                    vendor_id = entry[1]
                    break
            else:
                vendor_id = self._vendor_id

        if (ctx[0] != DirectoryContext.VENDOR or ctx[1] == vendor_id):
            ctx = (DirectoryContext.VENDOR, vendor_id)

        return self._parse_directory_entries(key_type, ctx, entries,
                                             self._COMMON_KEYS)

    #
    # 7.7.6 Module_Info entry
    #
    def _parse_module_directory(self, ctx, key_type, entries):
        # See explanation of Table 8 – Key ID allocations.
        for entry in entries:
            if entry[0] == (KeyType.SPECIFIER_ID.value. EntryType.IMMEDIATE):
                vendor_id = entry[1]
                break
        else:
            for entry in entries:
                if entry[0] == (KeyType.VENDOR.value, EntryType.IMMEDIATE):
                    vendor_id = entry[1]
                    break
            else:
                vendor_id = self._vendor_id

        if (ctx[0] != DirectoryContext.VENDOR or ctx[1] == vendor_id):
            ctx = (DirectoryContext.VENDOR, vendor_id)

        return self._parse_directory_entries(key_type, ctx, entries,
                                             self._COMMON_KEYS)

    #
    # 7.6.4 Feature directories
    #
    def _parse_feature_directory(self, ctx, key_type, entries):
        DEFINED_KEYS = {
            # name:  (key_type, available types of parser)
            KeyType.SPECIFIER_ID:   (EntryType.IMMEDIATE, ),
            KeyType.VERSION:        (EntryType.IMMEDIATE, ),
            KeyType.DEPENDENT_INFO: (EntryType.IMMEDIATE,
                                     EntryType.CSR_OFFSET,
                                     EntryType.LEAF,
                                     EntryType.DIRECTORY, ),
        }

        # Mandatory entries are required to decide directory context.
        for entry in entries:
            if entry[0] == (KeyType.SPECIFIER_ID.value, EntryType.IMMEDIATE):
                specifier_id = entry[1]
            elif entry[0] == (KeyType.VERSION.value, EntryType.IMMEDIATE):
                version = entry[1]
                break
        else:
            raise ValueError(
                'Mandatory entries are missing in feature directory.')
        ctx = (DirectoryContext.SPECIFIER, (specifier_id, version))

        keys = self._merge_common_keys(DEFINED_KEYS)

        return self._parse_directory_entries(key_type, ctx, entries, keys)

    #
    # 7.6.3 Unit directories
    #
    def _parse_unit_directory(self, ctx, key_type, entries):
        DEFINED_KEYS = {
            # name:  (key_type, available types of parser)
            KeyType.VENDOR:         (EntryType.IMMEDIATE,
                                     EntryType.LEAF,
                                     EntryType.DIRECTORY, ),
            KeyType.MODEL:          (EntryType.IMMEDIATE, ),
            KeyType.SPECIFIER_ID:   (EntryType.IMMEDIATE, ),
            KeyType.VERSION:        (EntryType.IMMEDIATE, ),
            KeyType.DEPENDENT_INFO: (EntryType.IMMEDIATE,
                                     EntryType.CSR_OFFSET,
                                     EntryType.LEAF,
                                     EntryType.DIRECTORY, ),
            KeyType.FEATURE:        (EntryType.DIRECTORY, ),
        }

        # Mandatory entries are required to decide directory context.
        for entry in entries:
            if entry[0] == (KeyType.SPECIFIER_ID.value, EntryType.IMMEDIATE):
                specifier_id = entry[1]
            elif entry[0] == (KeyType.VERSION.value, EntryType.IMMEDIATE):
                version = entry[1]
                break
        else:
            raise ValueError(
                'Mandatory entries are missing in unit directory.')
        ctx = (DirectoryContext.SPECIFIER, (specifier_id, version))

        keys = self._merge_common_keys(DEFINED_KEYS)

        return self._parse_directory_entries(key_type, ctx, entries, keys)

    #
    # 7.7.12 Dependent_Info entry
    #
    def _parse_dependent_info_directory(self, ctx, key_type, entries):
        # Mandatory entries are required to decide directory context.
        for entry in entries:
            if entry[0] == (KeyType.SPECIFIER_ID.value, EntryType.IMMEDIATE):
                specifier_id = entry[1]
            elif entry[0] == (KeyType.VERSION.value, EntryType.IMMEDIATE):
                version = entry[1]
                ctx = (DirectoryContext.SPECIFIER, (specifier_id, version))
                break
        else:
            # TODO: this is a work around. Precisely, need to decide according
            # to entries in parent directory voluntarily.
            pass

        return self._parse_directory_entries(key_type, ctx, entries,
                                             self._COMMON_KEYS)

    #
    # 7.6.2 Instance directories
    #
    def _parse_instance_directory(self, ctx, key_type, entries):
        DEFINED_KEYS = {
            # name:  (key_type, available types of parser)
            KeyType.VENDOR:         (EntryType.IMMEDIATE,
                                     EntryType.LEAF,
                                     EntryType.DIRECTORY, ),
            KeyType.KEYWORD:        (EntryType.LEAF, ),
            KeyType.FEATURE:        (EntryType.DIRECTORY, ),
            KeyType.INSTANCE:       (EntryType.DIRECTORY, ),
            KeyType.UNIT:           (EntryType.DIRECTORY, ),
            KeyType.MODEL:          (EntryType.IMMEDIATE, ),
            KeyType.DEPENDENT_INFO: (EntryType.DIRECTORY, ),
        }

        # Mandatory entries are required to decide directory context.
        for entry in entries:
            if entry[0] == (KeyType.KEYWORD, EntryType.IMMEDIATE):
                keyword = self._parse_leaf(entry[0][0], ctx, entry[1])
                break
        else:
            raise ValueError(
                'Mandatory entry is missing in instance directory.')

        ctx = (DirectoryContext.KEYWORD, keyword)

        keys = self._merge_common_keys(DEFINED_KEYS)

        return self._parse_directory_entries(key_type, ctx, entries, keys)

    def _parse_directory(self, key_type, ctx, entries):
        DIRECTORY_PARSERS = {
            KeyType.DESCRIPTOR:         self._parse_descriptor_directory,
            KeyType.BUS_DEPENDENT_INFO: self._parse_bus_dependent_directory,
            KeyType.VENDOR:             self._parse_vendor_directory,
            KeyType.MODULE:             self._parse_module_directory,
            KeyType.FEATURE:            self._parse_feature_directory,
            KeyType.UNIT:               self._parse_unit_directory,
            KeyType.DEPENDENT_INFO:     self._parse_dependent_info_directory,
            KeyType.INSTANCE:           self._parse_instance_directory,
        }

        return DIRECTORY_PARSERS[key_type](ctx, key_type, entries)

    def _parse_directory_entries(self, dir_key_type, ctx, entries, keys):
        TYPE_PARSERS = {
            EntryType.IMMEDIATE:    self.__parse_immediate,
            EntryType.CSR_OFFSET: lambda key_type, ctx, data: data,
            EntryType.LEAF:         self._parse_leaf,
            EntryType.DIRECTORY:    self._parse_directory,
        }
        EXTERNAL_HANDLES = {
            DirectoryContext.VENDOR:        self._vendor_dep_handles,
            DirectoryContext.SPECIFIER:     self._spec_dep_handles,
            DirectoryContext.BUS_DEPENDENT: self._bus_dep_handles,
            DirectoryContext.KEYWORD:       self._keyword_dep_handles,
        }
        info = []

        for entry in entries:
            key = entry[0]
            data = entry[1]
            parser = TYPE_PARSERS[key[1]]

            key_type = KeyType(key[0]) if KeyType.check_value(key[0]) else None

            if key_type in keys and key[1] in keys[key_type]:
                elem = [key_type.name, parser(key_type, ctx, data)]
            else:
                ctx_name, ctx_value = ctx
                if ctx_value not in EXTERNAL_HANDLES[ctx_name]:
                    elem = entry
                else:
                    for handle in EXTERNAL_HANDLES[ctx_name][ctx_value]:
                        elem = handle(key[0], key[1].name, data)
                        if elem:
                            break
                    else:
                        elem = entry

            info.append(elem)

        return info

    def parse_root_directory(self, bus_name, entries):
        DEFINED_KEYS = {
            # key_type:  available types of parser
            KeyType.BUS_DEPENDENT_INFO: (EntryType.IMMEDIATE,
                                         EntryType.CSR_OFFSET,
                                         EntryType.LEAF, ),
            KeyType.VENDOR:             (EntryType.IMMEDIATE,
                                         EntryType.LEAF,
                                         EntryType.DIRECTORY, ),
            KeyType.HARDWARE_VERSION:   (EntryType.IMMEDIATE, ),
            KeyType.MODULE:             (EntryType.LEAF,
                                         EntryType.DIRECTORY, ),
            KeyType.NODE_CAPABILITIES:  (EntryType.IMMEDIATE, ),
            KeyType.INSTANCE:           (EntryType.DIRECTORY, ),
            KeyType.UNIT:               (EntryType.DIRECTORY, ),
            KeyType.MODEL:              (EntryType.IMMEDIATE, ),
            KeyType.DEPENDENT_INFO:     (EntryType.DIRECTORY, ),
            # Node_Unique_ID was obsoleted.
        }

        # Mandatory entries are required to decide directory context.
        for entry in entries:
            if entry[0] == (KeyType.VENDOR.value, EntryType.IMMEDIATE):
                self._vendor_id = entry[1]
                break
        else:
            raise ValueError('Mandatory entry is missing in root directory.')
        ctx = (DirectoryContext.VENDOR, self._vendor_id)

        self._bus_name = bus_name

        keys = self._merge_common_keys(DEFINED_KEYS)

        return self._parse_directory_entries(KeyType.ROOT, ctx, entries, keys)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2019 Takashi Sakamoto


class FFMixerRegs():
    @classmethod
    def __generate_labels(cls, spec, category):
        labels = []
        for i in range(spec[category]):
            labels.append('{0}-{1}'.format(category, i + 1))
        return labels

    @classmethod
    def get_mixer_labels(cls, spec: dict):
        labels = cls.__generate_labels(spec, 'analog')
        labels += cls.__generate_labels(spec, 'spdif')
        labels += cls.__generate_labels(spec, 'adat')
        return labels

    @classmethod
    def get_mixer_src_labels(cls, spec: dict):
        labels = cls.__generate_labels(spec, 'analog')
        labels += cls.__generate_labels(spec, 'spdif')
        labels += cls.__generate_labels(spec, 'adat')
        labels += cls.__generate_labels(spec, 'stream')
        return labels

    @classmethod
    def calculate_src_offset(cls, spec: dict, dst: str, src: str):
        """
        Register layout.
             =+=========+=
              | inputs  | <= avail
        dst-0 +---------+
              | streams | <= avail
             =+=========+=
              | inputs  |
        dst-1 +---------+
              | streams |
             =+=========+=
        ...
             =+=========+=
              | inputs  |
        dst-n +---------+
              | streams |
             =+=========+=
        """

        dsts = cls.get_mixer_labels(spec)
        srcs = cls.get_mixer_src_labels(spec)
        if dst not in dsts:
            raise ValueError('Invalid argument for destination of mixer.')
        if src not in srcs:
            raise ValueError('Invalid argument for source of mixer.')

        offset = dsts.index(dst) * spec['avail'] * 2 * 4

        if src.find('stream-') == 0:
            offset += spec['avail'] * 4
            srcs = cls.__generate_labels(spec, 'stream')

        return offset + srcs.index(src) * 4



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.motu.motu_protocol_abstract import MotuProtocolAbstract

__all__ = ['MotuProtocolV3']


class MotuProtocolV3(MotuProtocolAbstract):
    def get_supported_sampling_rates(self):
        return self.SUPPORTED_SAMPLING_RATES_X4

    def get_sampling_rate(self):
        frames = self.read(0x0b14, 4)
        val = frames[2]
        return self.SUPPORTED_SAMPLING_RATES_X4[val]

    def set_sampling_rate(self, rate):
        frames = self.read(0x0b14, 4)
        frames[2] = self.SUPPORTED_SAMPLING_RATES_X4.index(rate)
        self.write(0x0b14, frames)

    def get_supported_clock_sources(self):
        sources = [
            self.CLOCK_INTERNAL,
            self.CLOCK_WORD_ON_BNC,
        ]

        flag = False
        mode = self.get_opt_iface_mode('in', 'A')
        if mode == 'ADAT':
            sources.append(self.CLOCK_ADAT_ON_OPT_A)
            flag = True
        elif mode == 'S/PDIF':
            sources.append(self.CLOCK_SPDIF_ON_OPT_A)

        mode = self.get_opt_iface_mode('in', 'B')
        if mode == 'ADAT':
            sources.append(self.CLOCK_ADAT_ON_OPT_B)
            flag = True
        elif mode == 'S/PDIF':
            sources.append(self.CLOCK_SPDIF_ON_OPT_B)

        if not flag:
            sources.append(self.CLOCK_SPDIF_ON_COAX)

        return sources

    def get_clock_source(self):
        data = self.read(0x0b14, 4)

        if data[3] == 0x00:
            return self.CLOCK_INTERNAL
        elif data[3] == 0x01:
            return self.CLOCK_WORD_ON_BNC
        elif data[3] == 0x10:
            return self.CLOCK_SPDIF_ON_COAX
        elif data[3] == 0x18:
            mode = self.get_opt_iface_mode('in', 'A')
            if mode == 'S/PDIF':
                return self.CLOCK_SPDIF_ON_OPT_A
            else:
                return self.CLOCK_ADAT_ON_OPT_A
        elif data[3] == 0x19:
            mode = self.get_opt_iface_mode('in', 'B')
            if mode == 'S/PDIF':
                return self.CLOCK_SPDIF_ON_OPT_B
            else:
                return self.CLOCK_ADAT_ON_OPT_B

        # TODO: signal from AES/EBU on XLR interface.

        return self.CLOCK_UNKNOWN

    def set_clock_source(self, src):
        data = self.read(0x0b14, 4)

        if src == self.CLOCK_INTERNAL:
            data[3] = 0x00
        elif src == self.CLOCK_WORD_ON_BNC:
            data[3] = 0x01
        elif src == self.CLOCK_SPDIF_ON_COAX:
            data[3] = 0x10
        elif src in (self.CLOCK_SPDIF_ON_OPT_A, self.CLOCK_ADAT_ON_OPT_A):
            mode = self.get_opt_iface_mode('in', 'A')
            if mode not in ('S/PDIF', 'ADAT'):
                raise OSError('Signal from this source is not available.')
            data[3] = 0x18
        elif src in (self.CLOCK_SPDIF_ON_OPT_B, self.CLOCK_ADAT_ON_OPT_B):
            mode = self.get_opt_iface_mode('in', 'B')
            if mode in ('S/PDIF', 'ADAT'):
                raise OSError('Signal from this source is not available.')
            data[3] = 0x19
        else:
            raise ValueError('Invalid argument for source of sampling clock.')

        # TODO: signal from AES/EBU on XLR interface

        self.write(0x0b14, data)

    # kind/direction/index/mask/shift
    OPT_IFACE_MODE_ATTRS = {
        'enabled': {
            'in': {
                # label: (pos, shift)
                'A': (3, 0),
                'B': (3, 1),
            },
            'out': {
                'A': (2, 0),
                'B': (2, 1),
            }
        },
        'no-adat': {
            'in': {
                'A': (1, 0),
                'B': (1, 4),
            },
            'out': {
                'A': (1, 2),
                'B': (1, 6),
            }
        },
    }

    def get_supported_opt_iface_indexes(self):
        return self.SUPPORTED_OPT_IFACE_INDEXES

    def get_opt_iface_mode(self, direction, index):
        frames = self.read(0x0c94, 4)

        pos, shift = self.OPT_IFACE_MODE_ATTRS['enabled'][direction][index]
        if frames[pos] & (1 << shift):
            # Enabled.
            pos, shift = self.OPT_IFACE_MODE_ATTRS['no-adat'][direction][index]
            if frames[pos] & (1 << shift):
                # S/PDIF
                idx = 1
            else:
                idx = 2
        else:
            # Disabled.
            idx = 0

        return self.SUPPORTED_OPT_IFACE_MODES[idx]

    def set_opt_iface_mode(self, direction, index, mode):
        idx = self.SUPPORTED_OPT_IFACE_MODES.index(mode)
        frames = self.read(0x0c94, 4)

        pos, shift = self.OPT_IFACE_MODE_ATTRS['enabled'][direction][index]
        frames[pos] & ~(1 << shift)
        if idx > 0:
            # Enabled.
            frames[pos] |= 1 << shift

            pos, shift = self.OPT_IFACE_MODE_ATTRS['no-adat'][direction][index]
            frames[pos] &= ~(1 << shift)
            if idx == 1:
                # S/PDIF.
                frames[pos] |= 1 << shift

        self.write(0x0c94, frames)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.ieee1394.config_rom_parser import Ieee1394ConfigRomParser

__all__ = ['TscmConfigRomParser']


class TscmConfigRomParser(Ieee1394ConfigRomParser):
    __OUI = 0x00022e

    def __init__(self):
        super().__init__()
        # FW-1884
        self.add_spec_dep_handle(self.__OUI, 0x800000, self.__handle_teac_keys)
        # FW-1082
        self.add_spec_dep_handle(self.__OUI, 0x800003, self.__handle_teac_keys)
        # FW-1804
        self.add_spec_dep_handle(self.__OUI, 0x800004, self.__handle_teac_keys)

    def __handle_teac_keys(self, key_id, type_name, data):
        if key_id == 0x02 and type_name == 'LEAF':
            content = data[8:].decode('US-ASCII') + '\0'
            return ['MODEL_NAME', content[:content.find('\0')]]
        return None

    def parse_rom(self, data):
        entries = super().parse_rom(data)
        return self.__parse_entries(entries['root-directory'])

    def __parse_entries(self, entries):
        # Typical layout.
        FIELDS = (
            ('VENDOR',              'vendor-id'),
            ('NODE_CAPABILITIES',   'node-capabilities'),
            ('EUI_64',              'guid'),
            ('UNIT',                None),
        )
        info = {}

        for i, field in enumerate(FIELDS):
            entry = entries[i]
            name, alt = field
            if entry[0] != name:
                raise OSError('Invalid format of config ROM.')
            if name == 'UNIT':
                # Check unit.
                entry = entry[1]
                if (entry[0] != ['SPECIFIER_ID', self.__OUI] or
                    entry[1][0] != 'VERSION' or
                        entry[2][0] != 'DEPENDENT_INFO'):
                    raise ValueError('Invalid data in unit directory.')
                info['model-version'] = entry[1][1]

                entry = entry[2][1]
                if (entry[0][0] != 'DESCRIPTOR' or
                        entry[1][0] != 'MODEL_NAME'):
                    raise ValueError(
                        'Invalid data in dependent info directory.')
                info['vendor-name'] = entry[0][1]
                info['model-name'] = entry[1][1]
            else:
                info[alt] = entry[1]

        return info



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from threading import Thread
from struct import unpack
from time import sleep

import gi
gi.require_version('GLib', '2.0')
gi.require_version('Hinawa', '3.0')
from gi.repository import GLib, Hinawa

from hinawa_utils.ta1394.config_rom_parser import Ta1394ConfigRomParser
from hinawa_utils.ta1394.general import AvcConnection
from hinawa_utils.ta1394.streamformat import AvcStreamFormatInfo

__all__ = ['OxfwUnit']


class OxfwUnit(Hinawa.SndUnit):
    def __init__(self, path):
        super().__init__()
        self.open(path)
        if self.get_property('type') != 4:
            raise ValueError('The character device is not for OXFW unit')

        ctx = GLib.MainContext.new()
        self.create_source().attach(ctx)
        self.__unit_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__unit_th = Thread(target=lambda d: d.run(), args=(self.__unit_dispatcher, ))
        self.__unit_th.start()

        node = self.get_node()
        ctx = GLib.MainContext.new()
        node.create_source().attach(ctx)
        self.__node_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__node_th = Thread(target=lambda d: d.run(), args=(self.__node_dispatcher, ))
        self.__node_th.start()

        parser = Ta1394ConfigRomParser()
        info = parser.parse_rom(self.get_node().get_config_rom())
        self.vendor_name = info['vendor-name']
        self.model_name = info['model-name']

        self.fcp = Hinawa.FwFcp()
        self.fcp.bind(self.get_node())

        self.hw_info = self._parse_hardware_info()
        self.supported_sampling_rates = self._parse_supported_sampling_rates()
        self.supported_stream_formats = self._parse_supported_stream_formats()

    def release(self):
        self.fcp.unbind()
        self.__unit_dispatcher.quit()
        self.__node_dispatcher.quit()
        self.__unit_th.join()
        self.__node_th.join()

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.release()

    def _parse_hardware_info(self):
        hw_info = {}

        req = Hinawa.FwReq()

        frames = bytearray(4)
        frames = req.transaction(self.get_node(),
                Hinawa.FwTcode.READ_QUADLET_REQUEST, 0xfffff0050000, 4, frames)
        hw_info['asic-type'] = 'FW{0:x}'.format(
            unpack('>H', frames[0:2])[0] >> 4)
        hw_info['firmware-version'] = '{0}.{1}'.format(frames[2], frames[3])

        frames = bytearray(4)
        frames = req.transaction(self.get_node(),
                Hinawa.FwTcode.READ_QUADLET_REQUEST, 0xfffff0090020, 4, frames)
        hw_info['asic-id'] = frames.decode('US-ASCII').rstrip('\0')

        return hw_info

    def _parse_supported_sampling_rates(self):
        sampling_rates = {}
        playback = []
        capture = []
        # Assume that PCM playback is available for all of models.
        for rate in AvcConnection.SAMPLING_RATES:
            if AvcConnection.ask_plug_signal_format(self.fcp, 'input', 0, rate):
                playback.append(rate)
        sleep(0.02)
        # PCM capture is not always available depending on models.
        for rate in AvcConnection.SAMPLING_RATES:
            if AvcConnection.ask_plug_signal_format(self.fcp, 'output', 0, rate):
                capture.append(rate)
        self._playback_only = (len(capture) == 0)
        for rate in AvcConnection.SAMPLING_RATES:
            if rate in playback or rate in capture:
                sampling_rates[rate] = True
        return sampling_rates

    def _parse_supported_stream_formats(self):
        supported_stream_formats = {}
        supported_stream_formats['playback'] = \
            AvcStreamFormatInfo.get_formats(self.fcp, 'input', 0)
        if len(supported_stream_formats['playback']) == 0:
            supported_stream_formats['playback'] = \
                self._assume_supported_stram_formats('input', 0)
            self._assumed = True
        else:
            self._assumed = False
        if not self._playback_only:
            supported_stream_formats['capture'] = \
                AvcStreamFormatInfo.get_formats(self.fcp, 'output', 0)
            if len(supported_stream_formats['capture']) == 0:
                supported_stream_formats['capture'] = \
                    self._assume_supported_stram_formats('output', 0)
        return supported_stream_formats

    def _assume_supported_stram_formats(self, direction, plug):
        assumed_stream_formats = []
        fmt = AvcStreamFormatInfo.get_format(self.fcp, 'input', 0)
        for rate, state in self.supported_sampling_rates.items():
            if state:
                assumed = {
                    'sampling-rate':    rate,
                    'rate-control':     fmt['rate-control'],
                    'formation':        fmt['formation']}
                assumed_stream_formats.append(assumed)
        return assumed_stream_formats

    def set_stream_formats(self, playback, capture):
        if playback not in self.supported_stream_formats['playback']:
            raise ValueError('Invalid argument for playback stream format')
        if capture:
            if self._playback_only:
                raise ValueError('This unit is playback only')
            if capture not in self.supported_stream_formats['capture']:
                raise ValueError('Invalid argument for capture stream format')
            if playback['sampling-rate'] != capture['sampling-rate']:
                raise ValueError(
                    'Sampling rate mis-match between playback and capture')
        if self._assumed:
            rate = playback['sampling-rate']
            AvcConnection.set_plug_signal_format(self.fcp, 'output', 0, rate)
            AvcConnection.set_plug_signal_format(self.fcp, 'input', 0, rate)
        else:
            AvcStreamFormatInfo.set_format(self.fcp, 'input', 0, playback)
            if not self._playback_only:
                AvcStreamFormatInfo.set_format(self.fcp, 'output', 0, capture)

    def get_current_stream_formats(self):
        playback = AvcStreamFormatInfo.get_format(self.fcp, 'input', 0)
        if not self._playback_only:
            capture = AvcStreamFormatInfo.get_format(self.fcp, 'output', 0)
        else:
            capture = None
        return {'playback': playback, 'capture': capture}



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from re import match
from struct import unpack

import gi
gi.require_version('Hinawa', '3.0')
from gi.repository import Hinawa

from hinawa_utils.bebob.maudio_protocol_abstract import MaudioProtocolAbstract

from hinawa_utils.ta1394.general import AvcConnection
from hinawa_utils.ta1394.ccm import AvcCcm
from hinawa_utils.ta1394.audio import AvcAudio

__all__ = ['MaudioProtocolNormal']


class MaudioProtocolNormal(MaudioProtocolAbstract):
    __IDS = (
        0x00000a,    # Ozonic
        0x010062,    # Firewire Solo
        0x010060,    # Firewire Audiophile
        0x010046,    # Firewire 410
        # Need information.
        #   NRV10
        #   ProFire Lightbridge
    )

    __LABELS = (
        {'inputs':  ('analog-1/2', 'analog-3/4', 'stream-1/2', 'stream-3/4'),
         'outputs': ('analog-1/2', 'analog-3/4'),
         'mixers':  ('mixer-1/2', 'mixer-3/4'),
         'meters':  ('analog-in-1', 'analog-in-2',
                     'analog-in-3', 'analog-in-4',
                     'stream-in-1', 'stream-in-2',
                     'stream-in-3', 'stream-in-4',
                     'analog-out-1', 'analog-out-2',
                     'analog-out-3', 'analog-out-4')},
        {'inputs':  ('analog-1/2', 'digital-1/2', 'stream-1/2', 'stream-3/4'),
         'outputs': ('analog-1/2', 'digital-1/2'),
         'mixers':  ('mixer-1/2', 'mixer-3/4'),
         'meters':  ('analog-in-1', 'analog-in-2',
                     'digital-in-1', 'digital-in-2',
                     'stream-in-1', 'stream-in-2',
                     'stream-in-3', 'stream-in-4',
                     'analog-out-1', 'analog-out-2',
                     'digital-out-1', 'digital-out-2')},
        {'inputs':  ('analog-1/2', 'digital-1/2',
                     'stream-1/2', 'stream-3/4', 'stream-5/6'),
         'outputs': ('analog-1/2', 'analog-3/4', 'digital-1/2'),
         'mixers':  ('mixer-1/2', 'mixer-3/4', 'mixer-5/6'),
         'meters':  ('analog-in-1', 'analog-in-2',
                     'digital-in-1', 'digital-in-2',
                     'analog-out-1', 'analog-out-2',
                     'analog-out-3', 'analog-out-4',
                     'digital-out-1', 'digital-out-2',
                     'headphone-out-1', 'headphone-out-2',
                     'aux-out-1', 'aux-out-2')},
        {'inputs':  ('analog-1/2', 'digital-1/2',
                     'stream-1/2', 'stream-3/4', 'stream-5/6', 'stream-7/8',
                     'stream-9/10'),
         'outputs': ('analog-1/2', 'analog-3/4', 'analog-5/6', 'analog-7/8',
                     'digital-1/2'),
         'mixers':  ('mixer-1/2', 'mixer-3/4', 'mixer-5/6', 'mixer-7/8',
                     'mixer-9/10'),
         'meters':  ('analog-in-1', 'analog-in-2',
                     'digital-in-1', 'digital-in-2',
                     'analog-out-1', 'analog-out-2',
                     'analog-out-3', 'analog-out-4',
                     'analog-out-5', 'analog-out-6',
                     'analog-out-7', 'analog-out-8',
                     'digital-out-1', 'digital-out-2',
                     'headphone-out-1', 'headphone-out-2',
                     'aux-out-1', 'aux-out-2')},
    )

    # = __LABELS['inputs']
    __INPUTS = (
        ((3, (1, 2)), (4, (1, 2)), (1, (1, 2)), (2, (1, 2))),
        ((1, (1, 2)), (2, (1, 2)), (4, (1, 2)), (3, (1, 2))),
        ((4, (1, 2)), (5, (1, 2)), (1, (1, 2)), (2, (1, 2)), (3, (1, 2))),
        ((3, (1, 2)), (4, (1, 2)),
         (2, (1, 2)), (1, (1, 2)), (1, (3, 4)), (1, (5, 6)), (1, (7, 8))),
    )

    # = __LABELS['inputs']
    __AUX__INPUTS = (
        (),
        (),
        ((9, (1, 2)), (10, (1, 2)), (6, (1, 2)), (7, (1, 2)), (8, (1, 2))),
        ((7, (1, 2)), (8, (1, 2)), (9, (1, 2)), (6, (1, 2)),
         (5, (1, 2)), (5, (3, 4)), (5, (5, 6)), (5, (7, 8))),
    )

    __AUX_OUTPUT = (
        None,
        None,
        11,
        9,
    )

    # = __LABELS['inputs']
    __MIXER_SOURCES = (
        ((2, (1, 2)), (3, (1, 2)), (0, (1, 2)), (1, (1, 2))),
        ((0, (1, 2)), (1, (1, 2)), (3, (1, 2)), (2, (1, 2))),
        ((3, (1, 2)), (4, (1, 2)), (0, (1, 2)), (1, (1, 2)), (2, (1, 2))),
        ((2, (1, 2)), (3, (1, 2)),
         (1, (1, 2)), (0, (1, 2)), (0, (3, 4)), (0, (5, 6)),
         (0, (7, 8))),
    )

    # = __LABELS['mixers']
    __MIXERS = (
        ((1, (1, 2)), (1, (1, 2))),
        ((1, (1, 2)), (1, (3, 4))),
        ((1, (1, 2)), (2, (1, 2)), (3, (1, 2))),
        ((1, (1, 2)), (1, (3, 4)), (1, (5, 6)), (1, (7, 8)), (1, (9, 10))),
    )

    # = __LABELS['outputs']
    __OUTPUT_SOURCES = (
        (),
        (),
        (1, 2, 3),
        (2, 3, 4, 5, 6),
    )

    # = __LABELS['outputs']
    __OUTPUTS = (
        (),
        (),
        ((12, (1, 2)), (13, (1, 2)), (14, (1, 2))),
        ((10, (1, 2)), (11, (1, 2)), (12, (1, 2)), (13, (1, 2)), (14, (1, 2))),
    )

    __HP_SOURCES = (
        (),
        (),
        ((4, (0, 1, 2, 3)), ),
        ((7, (2, 3, 4, 5, 6, 7)), ),
    )

    __HP_OUTS = (
        (),
        (),
        ((15, (1, 2)), ),
        ((15, (1, 2)), ),
    )

    __METERS = (
        48,
        52,
        60,
        76,
    )

    __CLOCKS = (
        {},
        {'Internal':    AvcCcm.get_subunit_signal_addr('music', 0, 1),
         'S/PDIF':      AvcCcm.get_unit_signal_addr('external', 1)},
        {'Internal':    AvcCcm.get_subunit_signal_addr('music', 0, 1),
         'S/PDIF':      AvcCcm.get_unit_signal_addr('external', 2)},
        {'Internal':    AvcCcm.get_subunit_signal_addr('music', 0, 1),
         'S/PDIF':      AvcCcm.get_unit_signal_addr('external', 2),
         'ADAT':        AvcCcm.get_unit_signal_addr('external', 3)},
    )

    def __init__(self, unit, debug):
        if unit.model_id not in self.__IDS:
            raise OSError('Not supported')

        super().__init__(unit, debug)

        index = self.__IDS.index(unit.model_id)
        self.labels = self.__LABELS[index]
        self.mixers = self.__MIXERS[index]
        self.__inputs = self.__INPUTS[index]
        self.__aux_inputs = self.__AUX__INPUTS[index]
        self.__aux_output = self.__AUX_OUTPUT[index]
        self.__mixer_sources = self.__MIXER_SOURCES[index]
        self.__output_sources = self.__OUTPUT_SOURCES[index]
        self.__outputs = self.__OUTPUTS[index]
        self.__hp_sources = self.__HP_SOURCES[index]
        self.__hp_outs = self.__HP_OUTS[index]
        self.__meters = self.__METERS[index]
        self.__clocks = self.__CLOCKS[index]

    def _refer_fb_data(self, targets, index, ch):
        if index >= len(targets):
            raise ValueError('Invalid argument for function block index')
        if ch >= len(targets[index][1]):
            raise ValueError('Invalid argument for channel number')
        fb = targets[index][0]
        ch = targets[index][1][ch]
        return (fb, ch)

    def _set_volume(self, targets, index, ch, db):
        fb, ch = self._refer_fb_data(targets, index, ch)
        data = AvcAudio.build_data_from_db(db)
        AvcAudio.set_feature_volume_state(self.unit.fcp, 0, 'current', fb, ch,
                                          data)

    def _get_volume(self, targets, index, ch):
        fb, ch = self._refer_fb_data(targets, index, ch)
        data = AvcAudio.get_feature_volume_state(self.unit.fcp, 0, 'current',
                                                 fb, ch)
        return AvcAudio.parse_data_to_db(data)

    def get_input_labels(self):
        return self.labels['inputs']

    def _refer_input_data(self, target):
        if target not in self.labels['inputs']:
            raise ValueError('Invalid argument for input')
        return self.labels['inputs'].index(target)

    def set_input_gain(self, target, ch, db):
        index = self._refer_input_data(target)
        self._set_volume(self.__inputs, index, ch, db)

    def get_input_gain(self, target, ch):
        index = self._refer_input_data(target)
        return self._get_volume(self.__inputs, index, ch)

    def get_input_balance_labels(self):
        labels = []
        for label in self.labels['inputs']:
            if label.find('stream-') == 0:
                continue
            labels.append(label)
        return labels

    def set_input_balance(self, target, ch, balance):
        index = self._refer_input_data(target)
        fb, ch = self._refer_fb_data(self.__inputs, index, ch)
        data = AvcAudio.build_data_from_db(balance)
        AvcAudio.set_feature_lr_state(self.unit.fcp, 0, 'current', fb, ch,
                                      data)

    def get_input_balance(self, target, ch):
        index = self._refer_input_data(target)
        fb, ch = self._refer_fb_data(self.__inputs, index, ch)
        data = AvcAudio.get_feature_lr_state(self.unit.fcp, 0, 'current',
                                             fb, ch)
        return AvcAudio.parse_data_to_db(data)

    def get_output_labels(self):
        if len(self.__outputs) == 0:
            return ()
        return self.labels['outputs']

    def _refer_out_data(self, target):
        if target not in self.labels['outputs']:
            raise ValueError('Invalid argument for output')
        return self.labels['outputs'].index(target)

    def set_output_volume(self, target, ch, db):
        index = self._refer_out_data(target)
        self._set_volume(self.__outputs, index, ch, db)

    def get_output_volume(self, target, ch):
        index = self._refer_out_data(target)
        return self._get_volume(self.__outputs, index, ch)

    def set_aux_volume(self, ch, db):
        if ch > 2:
            raise ValueError('Invalid argument for master channel')
        fb = self.__aux_output
        data = AvcAudio.build_data_from_db(db)
        AvcAudio.set_feature_volume_state(self.unit.fcp, 0, 'current', fb, ch,
                                          data)

    def get_aux_volume(self, ch):
        if ch > 2:
            raise ValueError('Invalid argument for master channel')
        fb = self.__aux_output
        data = AvcAudio.get_feature_volume_state(self.unit.fcp, 0, 'current',
                                                 fb, ch)
        return AvcAudio.parse_data_to_db(data)

    def set_aux_balance(self, ch, balance):
        if ch > 2:
            raise ValueError('Invalid argument for master channel')
        fb = self.__aux_output
        data = AvcAudio.build_data_from_db(balance)
        ch += 1
        AvcAudio.set_feature_lr_state(self.unit.fcp, 0, 'current', fb, ch,
                                      data)

    def get_aux_balance(self, ch):
        if ch > 2:
            raise ValueError('Invalid argument for master channel')
        fb = self.__aux_output
        ch += 1
        data = AvcAudio.get_feature_lr_state(self.unit.fcp, 0, 'current',
                                             fb, ch)
        return AvcAudio.parse_data_to_db(data)

    def get_headphone_labels(self):
        labels = []
        for i in range(len(self.__hp_outs)):
            labels.append('headphone-{0}/{1}'.format(i * 2 + 1, i * 2 + 2))
        return labels

    def _refer_hp_data(self, target):
        matches = match('^headphone-([0-9]*)/([0-9]*)$', target)
        if not matches:
            raise ValueError('Invalid argument for headphone')
        left = int(matches.group(1)) // 2
        right = int(matches.group(2)) // 2
        if right != left + 1:
            raise ValueError('Invalid argument for headphone')
        return left

    def set_headphone_volume(self, target, ch, db):
        index = self._refer_hp_data(target)
        self._set_volume(self.__hp_outs, index, ch, db)

    def get_headphone_volume(self, target, ch):
        index = self._refer_hp_data(target)
        return self._get_volume(self.__hp_outs, index, ch)

    def get_aux_input_labels(self):
        if not self.__aux_output:
            return ()
        return self.labels['inputs']

    def set_aux_input(self, target, ch, db):
        index = self._refer_input_data(target)
        self._set_volume(self.__aux_inputs, index, ch, db)

    def get_aux_input(self, target, ch):
        index = self._refer_input_data(target)
        return self._get_volume(self.__aux_inputs, index, ch)

    def get_mixer_labels(self):
        return self.labels['mixers']

    def get_mixer_source_labels(self):
        return self.labels['inputs']

    def _refer_mixer_data(self, target, source):
        if source not in self.labels['inputs']:
            raise ValueError('Invalid argument for mixer input')
        if target not in self.labels['mixers']:
            raise ValueError('Invalid argument for mixer output')
        input = self.labels['inputs'].index(source)
        in_fb = self.__mixer_sources[input][0]
        in_ch = self.__mixer_sources[input][1][0]   # Use left channel.
        mixer = self.labels['mixers'].index(target)
        out_fb = self.mixers[mixer][0]
        out_ch = self.mixers[mixer][1][0]  # Use left channel.
        return (in_fb, in_ch, out_fb, out_ch)

    def set_mixer_routing(self, target, source, enable):
        in_fb, in_ch, out_fb, out_ch = self._refer_mixer_data(target, source)
        if enable:
            data = (0x00, 0x00)
        else:
            data = (0x80, 0x00)
        AvcAudio.set_processing_mixer_state(self.unit.fcp, 0, 'current',
                                            out_fb, in_fb, in_ch, out_ch, data)

    def get_mixer_routing(self, target, source):
        in_fb, in_ch, out_fb, out_ch = self._refer_mixer_data(target, source)
        data = AvcAudio.get_processing_mixer_state(self.unit.fcp, 0, 'current',
                                                   out_fb, in_fb, in_ch, out_ch)
        return data[0] == 0x00 and data[1] == 0x00

    def get_headphone_source_labels(self, target):
        labels = []
        if len(self.__hp_sources) > 0:
            for mixer in self.labels['mixers']:
                labels.append(mixer)
            if self.__aux_output:
                labels.append("aux-1/2")
        return labels

    def set_headphone_source(self, target, source):
        index = self._refer_hp_data(target)
        if source in self.labels['mixers']:
            ch = self.labels['mixers'].index(source)
        elif source.find('aux') == 0:
            ch = len(self.labels['mixers'])
        else:
            raise ValueError('Invalid argument for output target')
        fb = self.__hp_sources[index][0]
        value = self.__hp_sources[index][1][ch]
        AvcAudio.set_selector_state(self.unit.fcp, 0, 'current', fb, value)

    def get_headphone_source(self, target):
        index = self._refer_hp_data(target)
        fb = self.__hp_sources[index][0]
        value = AvcAudio.get_selector_state(self.unit.fcp, 0, 'current', fb)
        ch = self.__hp_sources[index][1][value]
        if ch < len(self.labels['mixers']):
            return self.labels['mixers'][ch]
        return 'aux-1/2'

    def get_output_source_labels(self, target):
        index = self._refer_out_data(target)
        labels = []
        labels.append(self.labels['mixers'][index])
        if self.__aux_output:
            labels.append("aux-1/2")
        return labels

    def set_output_source(self, target, source):
        index = self._refer_out_data(target)
        if source in self.labels['mixers'][index]:
            value = 0
        elif source.find('aux') == 0:
            value = 1
        else:
            raise ValueError('Invalid argument for output target')
        fb = self.__output_sources[index]
        AvcAudio.set_selector_state(self.unit.fcp, 0, 'current', fb, value)

    def get_output_source(self, target):
        index = self._refer_out_data(target)
        fb = self.__output_sources[index]
        value = AvcAudio.get_selector_state(self.unit.fcp, 0, 'current', fb)
        if value == 1 and self.__aux_output:
            return 'aux-1/2'
        return self.labels['mixers'][index]

    # 0x0000ffff - 0x7fffffff
    # db = 20 * log10(vol / 0x80000000)
    # vol = 0, then db = -144.0
    # may differs analog-in and the others.
    def get_meters(self):
        labels = self.labels['meters']
        meters = {}
        req = Hinawa.FwReq()
        frames = [0] * 256
        data = req.transaction(self.unit.get_node(),
                Hinawa.FwTcode.READ_BLOCK_REQUEST, self._ADDR_FOR_METERING,
                self.__meters, frames)
        for i, name in enumerate(labels):
            meters[name] = unpack('>I', data[i * 4:(i + 1) * 4])[0]
        if len(data) > len(labels) * 4:
            meters['rotery-0'] = data[-3] & 0x0f
            meters['rotery-1'] = (data[-3] & 0xf0) >> 4
            meters['rotery-2'] = 0
            meters['switch-0'] = (data[-4] & 0xf0) >> 4
            meters['switch-1'] = data[-4] & 0x0f
            meters['rate'] = AvcConnection.SAMPLING_RATES[data[-2]]
            meters['sync'] = data[-1] & 0x0f
        return meters

    def get_clock_source_labels(self):
        return self.__clocks.keys()

    def set_clock_source(self, src):
        if self.unit.get_property('streaming'):
            raise ValueError('Packet streaming already runs.')
        dst = AvcCcm.get_subunit_signal_addr('music', 0, 1)
        addr = self.__clocks[src]
        AvcCcm.set_signal_source(self.unit.fcp, addr, dst)

    def get_clock_source(self):
        dst = AvcCcm.get_subunit_signal_addr('music', 0, 1)
        curr = AvcCcm.get_signal_source(self.unit.fcp, dst)
        for name, addr in self.__clocks.items():
            if AvcCcm.compare_addrs(curr, AvcCcm.parse_signal_addr(addr)):
                return name

    def get_sampling_rate(self):
        in_rate = AvcConnection.get_plug_signal_format(self.unit.fcp, 'input',
                                                       0)
        out_rate = AvcConnection.get_plug_signal_format(self.unit.fcp,
                                                        'output', 0)
        if in_rate != out_rate:
            raise OSError('Unexpected state of the unit: {0}!={1}'.format(
                in_rate, out_rate))
        return in_rate



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.motu.motu_protocol_abstract import MotuProtocolAbstract

__all__ = ['MotuProtocolV1']


class MotuProtocolV1(MotuProtocolAbstract):
    SUPPORTED_MODELS = ('828', '896')

    # direction/mask/shift
    OPT_IFACE_MODE_ATTRS = {
        'in':   ((2, 0x80, 3), (3, 0x80, 3)),
        'out':  ((2, 0x40, 2), (3, 0x40, 2))
    }

    PACKET_SIZE_ATTRS = {
        'in':   (3, 0x80, 3),
        'out':  (3, 0x40, 2),
    }

    def get_supported_sampling_rates(self):
        if self._unit.name == '828':
            return self.SUPPORTED_SAMPLING_RATES_X1
        elif self._unit.name == '896':
            return self.SUPPORTED_SAMPLING_RATES_X2

    def get_sampling_rate(self):
        frames = self.read(0x0b00, 4)
        if frames[3] & 0x04:
            rate = 48000
        else:
            rate = 44100
        if frames[3] & 0x08:
            rate * 2
        return rate

    def set_sampling_rate(self, rate):
        frames = self.read(0x0b00, 4)
        frames[0] &= ~0x0c
        if rate % 48000 == 0:
            frames[3] |= 0x04
        if rate > 48000:
            frames[3] |= 0x08
        self.write(0x0b00, frames)

    def get_supported_clock_sources(self):
        sources = [
            self.CLOCK_INTERNAL,
            self.CLOCK_ADAT_ON_DSUB,
        ]
        if self.get_opt_iface_mode('in', 'A') == 'S/PDIF':
            sources.append(self.CLOCK_SPDIF_ON_OPT)
        else:
            sources.append(self.CLOCK_ADAT_ON_OPT)
            sources.append(self.CLOCK_SPDIF_ON_COAX)

        if self._unit.name == '896':
            sources.append(self.CLOCK_WORD_ON_BNC)

        return sources

    def get_clock_source(self):
        frames = self.read(0x0b00, 4)

        if frames[3] & 0x01:
            if frames[3] & 0x20:
                return self.CLOCK_ADAT_ON_OPT
            else:
                return self.CLOCK_ADAT_ON_DSUB
        elif frames[3] & 0x02:
            if frames[2] & 0x80:
                return self.CLOCK_SPDIF_ON_OPT
            else:
                return self.CLOCK_SPDIF_ON_COAX
        else:
            return self.CLOCK_INTERNAL

    def set_clock_source(self, source):
        frames = self.read(0x0b00, 4)
        frames[2] &= ~0x80
        frames[3] &= ~0x23

        if source != self.CLOCK_ADAT_ON_DSUB:
            frames[3] |= 0x20

        if source in (self.CLOCK_ADAT_ON_DSUB, self.CLOCK_ADAT_ON_OPT):
            frames[3] |= 0x01
        elif source == self.CLOCK_SPDIF_ON_OPT:
            frames[3] |= 0x02
        elif source == self.CLOCK_SPDIF_ON_COAX:
            if frames[2] & 0x80:
                raise ValueError('This mode is currently unavailable.')
            frames[2] |= 0x80
            frames[3] |= 0x02
        elif source == self.CLOCK_WORD_ON_BNC:
            frames[3] |= 0x10

        self.write(0x0b00, frames)

    def get_supported_opt_iface_indexes(self):
        return self.SUPPORTED_OPT_IFACE_INDEXES[0:1]

    def get_opt_iface_mode(self, direction, index):
        frames = self.read(0x0b00, 4)

        pos, mask, shift = self.OPT_IFACE_MODE_ATTRS[direction][0]
        if frames[pos] & mask:
            return self.SUPPORTED_OPT_IFACE_MODES[1]
        else:
            # Need to check the size of data block, sigh...
            frames = self.read(0x0b10, 4)

            pos, mask, shift = self.OPT_IFACE_MODE_ATTRS[direction][1]
            if frames[pos] & mask:
                return self.SUPPORTED_OPT_IFACE_MODES[0]
            else:
                return self.SUPPORTED_OPT_IFACE_MODES[2]

    def set_opt_iface_mode(self, direction, index, mode):
        frames = self.read(0x0b00, 4)
        frames[0] &= 0xc0

        pos, mask, shift = self.OPT_IFACE_MODE_ATTRS[direction][0]
        if self.SUPPORTED_OPT_IFACE_MODES.index(mode) == 1:
            frames[pos] |= 1 << shift

        self.write(0x0b00, frames)

        pos, mask, shift = self.PACKET_SIZE_ATTRS[direction]
        frames = self.read(0x0b10, 4)
        frames[pos] &= ~mask
        if mode != 'ADAT':
            frames[pos] |= 1 << shift

        self.write(0x0b10, frames)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from threading import Thread

import gi
gi.require_version('GLib', '2.0')
gi.require_version('Hinawa', '3.0')
from gi.repository import GLib, Hinawa

from hinawa_utils.efw.transactions import EftInfo
from hinawa_utils.efw.transactions import EftHwctl
from hinawa_utils.efw.transactions import EftPhysOutput
from hinawa_utils.efw.transactions import EftPhysInput
from hinawa_utils.efw.transactions import EftPlayback
from hinawa_utils.efw.transactions import EftMonitor
from hinawa_utils.efw.transactions import EftIoconf
from math import log10

__all__ = ['EfwUnit']


class EfwUnit(Hinawa.SndEfw):
    def __init__(self, path):
        super().__init__()
        self.open(path)

        ctx = GLib.MainContext.new()
        self.create_source().attach(ctx)
        self.__unit_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__unit_th = Thread(target=lambda d: d.run(), args=(self.__unit_dispatcher, ))
        self.__unit_th.start()

        node = self.get_node()
        ctx = GLib.MainContext.new()
        node.create_source().attach(ctx)
        self.__node_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__node_th = Thread(target=lambda d: d.run(), args=(self.__node_dispatcher, ))
        self.__node_th.start()

        self.info = EftInfo.get_spec(self)
        self._fixup_info()

    def release(self):
        self.__unit_dispatcher.quit()
        self.__node_dispatcher.quit()
        self.__unit_th.join()
        self.__node_th.join()

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.release()

    def _fixup_info(self):
        # Mapping for channels on tx stream is supported by Onyx1200F only.
        if self.info['model'] == 'Onyx1200F':
            self.info['features']['tx-mapping'] = True
        else:
            self.info['features']['tx-mapping'] = False

        # S/PDIF on coaxial interfae is available in Onyx400F.
        if self.info['model'] == 'Onyx400F':
            self.info['features']['spdif-coax'] = True

        # Nominal level of input/output is always supported by AudioFire series.
        if self.info['model'].find('Audiofire') == 0:
            self.info['features']['nominal-input'] = True
            self.info['features']['nominal-output'] = True

    @staticmethod
    def _calcurate_vol_from_db(db):
        if db <= -144.0:
            return 0x00000000
        else:
            return int(0x01000000 * pow(10, db / 20))

    @staticmethod
    def _calcurate_vol_to_db(vol):
        if vol == 0:
            return -144.0
        else:
            return 20 * log10(vol / 0x01000000)

    def get_metering(self):
        return EftInfo.get_metering(self)

    def set_clock_state(self, rate, src):
        EftHwctl.set_clock(self, rate, src, 0)

    def get_clock_state(self):
        return EftHwctl.get_clock(self)

    def get_box_state_labels(self, name):
        if name not in EftHwctl.SUPPORTED_BOX_STATES:
            raise ValueError('Invalid argument for name of box state')
        return EftHwctl.SUPPORTED_BOX_STATES[name]

    def set_box_states(self, name, state):
        states = EftHwctl.get_box_states(self)
        states[name] = state
        EftHwctl.set_box_states(self, states)

    def get_box_states(self):
        states = {}
        state_all = EftHwctl.get_box_states(self)
        if self.info['features']['spdif-coax'] or \
           self.info['features']['spdif-opt']:
            states['spdif-pro'] = state_all['spdif-pro']
            states['spdif-non-audio'] = state_all['spdif-non-audio']
        if self.info['model'] == 'Onyx1200F':
            states['control-room'] = state_all['control-room']
            states['output-level-bypass'] = state_all['output-level-bypass']
            states['metering-mode-in'] = state_all['metering-mode-in']
            states['metering-mode-out'] = state_all['metering-mode-out']
        if self.info['features']['soft-clip']:
            states['soft-clip'] = state_all['soft-clip']
        if self.info['features']['robot-hex-input']:
            states['robot-hex-input'] = state_all['robot-hex-input']
        if self.info['features']['robot-battery-charge']:
            states['robot-battery-charge'] = state_all['robot-battery-charge']
        states['internal-multiplexer'] = state_all['internal-multiplexer']
        return states

    def set_phys_out_gain(self, ch, db):
        if ch >= len(self.info['phys-outputs']):
            raise ValueError('Invalid argument for physical output channel')
        vol = self._calcurate_vol_from_db(db)
        EftPhysOutput.set_param(self, 'gain', ch, vol)

    def get_phys_out_gain(self, ch):
        if ch >= len(self.info['phys-outputs']):
            raise ValueError('Invalid argument for physical output channel')
        vol = EftPhysOutput.get_param(self, 'gain', ch)
        return self._calcurate_vol_to_db(vol)

    def set_phys_out_mute(self, ch, val):
        if ch >= len(self.info['phys-outputs']):
            raise ValueError('Invalid argument for physical output channel')
        EftPhysOutput.set_param(self, 'mute', ch, val)

    def get_phys_out_mute(self, ch):
        if ch >= len(self.info['phys-outputs']):
            raise ValueError('Invalid argument for physical output channel')
        return EftPhysOutput.get_param(self, 'mute', ch)

    def set_phys_out_nominal(self, ch, val):
        if not self.info['features']['nominal-output']:
            raise RuntimeError('Not supported by this model')
        if ch >= len(self.info['phys-outputs']):
            raise ValueError('Invalid argument for physical output channel')
        EftPhysOutput.set_param(self, 'nominal', ch, val)

    def get_phys_out_nominal(self, ch):
        if not self.info['features']['nominal-output']:
            raise RuntimeError('Not supported by this model')
        if ch >= len(self.info['phys-outputs']):
            raise ValueError('Invalid argument for physical output channel')
        return EftPhysOutput.get_param(self, 'nominal', ch)

    def set_phys_in_nominal(self, ch, val):
        if not self.info['features']['nominal-input']:
            raise RuntimeError('Not supported by this model')
        if ch >= len(self.info['phys-inputs']):
            raise ValueError('Invalid argument for physical input channel')
        EftPhysInput.set_param(self, 'nominal', ch, val)

    def get_phys_in_nominal(self, ch):
        if not self.info['features']['nominal-input']:
            raise RuntimeError('Not supported by this model')
        if ch >= len(self.info['phys-inputs']):
            raise ValueError('Invalid argument for physical input channel')
        return EftPhysInput.get_param(self, 'nominal', ch)

    def set_playback_gain(self, ch, db):
        if ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        vol = self._calcurate_vol_from_db(db)
        EftPlayback.set_param(self, 'gain', ch, vol)

    def get_playback_gain(self, ch):
        if ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        vol = EftPlayback.get_param(self, 'gain', ch)
        return self._calcurate_vol_to_db(vol)

    def set_playback_mute(self, ch, val):
        if ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        EftPlayback.set_param(self, 'mute', ch, val)

    def get_playback_mute(self, ch):
        if ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        return EftPlayback.get_param(self, 'mute', ch)

    def set_playback_solo(self, ch, val):
        if ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        EftPlayback.set_param(self, 'solo', ch, val)

    def get_playback_solo(self, ch):
        if ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        return EftPlayback.get_param(self, 'solo', ch)

    def set_monitor_gain(self, in_ch, out_ch, db):
        if in_ch >= self.info['capture-channels']:
            raise ValueError('Invalid argument for capture channel')
        if out_ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        vol = self._calcurate_vol_from_db(db)
        EftMonitor.set_param(self, 'gain', in_ch, out_ch, vol)

    def get_monitor_gain(self, in_ch, out_ch):
        if in_ch >= self.info['capture-channels']:
            raise ValueError('Invalid argument for capture channel')
        if out_ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        vol = EftMonitor.get_param(self, 'gain', in_ch, out_ch)
        return self._calcurate_vol_to_db(vol)

    def set_monitor_mute(self, in_ch, out_ch, val):
        if in_ch >= self.info['capture-channels']:
            raise ValueError('Invalid argument for capture channel')
        if out_ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        EftMonitor.set_param(self, 'mute', in_ch, out_ch, val)

    def get_monitor_mute(self, in_ch, out_ch):
        if in_ch >= self.info['capture-channels']:
            raise ValueError('Invalid argument for capture channel')
        if out_ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        return EftMonitor.get_param(self, 'mute', in_ch, out_ch)

    def set_monitor_solo(self, in_ch, out_ch, val):
        if in_ch >= self.info['capture-channels']:
            raise ValueError('Invalid argument for capture channel')
        if out_ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        EftMonitor.set_param(self, 'solo', in_ch, out_ch, val)

    def get_monitor_solo(self, in_ch, out_ch):
        if in_ch >= self.info['capture-channels']:
            raise ValueError('Invalid argument for capture channel')
        if out_ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        return EftMonitor.get_param(self, 'solo', in_ch, out_ch)

    def set_monitor_pan(self, in_ch, out_ch, val):
        if in_ch >= self.info['capture-channels']:
            raise ValueError('Invalid argument for capture channel')
        if out_ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        EftMonitor.set_param(self, 'pan', in_ch, out_ch, val)

    def get_monitor_pan(self, in_ch, out_ch):
        if in_ch >= self.info['capture-channels']:
            raise ValueError('Invalid argument for capture channel')
        if out_ch >= self.info['playback-channels']:
            raise ValueError('Invalid argument for playback channel')
        return EftMonitor.get_param(self, 'pan', in_ch, out_ch)

    def get_control_room_source_labels(self):
        labels = []
        for i in range(1, self.info['playback-channels'], 2):
            labels.append('mixer-{0}/{1}'.format(i, i + 1))
        return labels

    def set_control_room_mirroring(self, source):
        if not self.info['features']['control-room-mirroring']:
            raise RuntimeError('Not supported by this model')
        labels = self.get_control_room_source_labels()
        if source not in labels:
            raise ValueError('Invalid argument for source')
        val = labels.index(source) * 2
        EftIoconf.set_control_room_mirroring(self, val)

    def get_control_room_mirroring(self):
        if not self.info['features']['control-room-mirroring']:
            raise RuntimeError('Not supported by this model')
        val = EftIoconf.get_control_room_mirroring(self)
        labels = self.get_control_room_source_labels()
        return labels[val // 2]

    def get_digital_input_mode_labels(self):
        labels = []
        for mode in EftIoconf.DIGITAL_INPUT_MODES:
            if mode in self.info['features'] and self.info['features'][mode]:
                labels.append(mode)
        return labels

    def set_digital_input_mode(self, mode):
        if not self.info['features'][mode]:
            raise RuntimeError('Not supported by this model')
        EftIoconf.set_digital_input_mode(self, mode)

    def get_digital_input_mode(self):
        return EftIoconf.get_digital_input_mode(self)

    def set_phantom_powering(self, state):
        if not self.info['features']['phantom-powering']:
            raise RuntimeError('Not supported by this model')
        EftIoconf.set_phantom_powering(self, state)

    def get_phantom_powering(self):
        if not self.info['features']['phantom-powering']:
            raise RuntimeError('Not supported by this model')
        return EftIoconf.get_phantom_powering(self)

    def set_stream_mapping(self, rx_maps, tx_maps):
        if not rx_maps and self.info['features']['rx-mapping']:
            raise RuntimeError('Not supported by this model')
        if not tx_maps and self.info['features']['tx-mapping']:
            raise RuntimeError('Not supported by this model')
        EftIoconf.set_stream_mapping(self, rx_maps, tx_maps)

    def get_stream_mapping(self):
        if not self.info['features']['rx-mapping'] and \
           not self.info['features']['tx-mapping']:
            raise RuntimeError('Not supported by this model')
        return EftIoconf.get_stream_mapping(self)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2019 Takashi Sakamoto

from enum import Enum
from struct import pack, unpack

import gi
gi.require_version('Hinawa', '3.0')
from gi.repository import Hinawa

from hinawa_utils.ta1394.general import AvcGeneral

__all__ = ['MicCmd', 'InputCmd', 'OutputCmd', 'MixerCmd', 'DisplayCmd',
           'KnobCmd']


class VendorCmd(Enum):
    MIC_POLARITY = 0x00
    IN_LEVEL = 0x01
    IN_ATTR = 0x02
    MIC_POWER = 0x03
    OUT_ATTR = 0x04
    IN_GAIN = 0x05
    HW_STATUS = 0x07
    # Unknown: 0x08
    OUT_MUTE = 0x09
    IN_SRC = 0x0c
    MIXER_SRC = 0x10
    OUT_SRC = 0x11
    DISPLAY_OVERHOLD = 0x13
    DISPLAY_CLEAR = 0x14
    OUT_VOLUME = 0x15
    # Out mute is engaged for speaker: 0x16
    # Out mute is engaged for headphone: 0x17
    # Out mute is not engaged for speaker: 0x18
    # Out mute is not engaged for headphone: 0x19
    DISPLAY_TARGET = 0x1b
    IN_CLICKLESS = 0x1e
    DISPLAY_FOLLOW = 0x22


class ApogeeProtocol():
    __OUI = (0x00, 0x03, 0xdb)

    @classmethod
    def __command_req(cls, fcp, cmd, args, func_name):
        funcs = {
            'set':  AvcGeneral.set_vendor_dependent,
            'get':  AvcGeneral.get_vendor_dependent,
        }

        data = bytearray(4)
        data[0] = 0x50    # P
        data[1] = 0x43    # C
        data[2] = 0x4d    # M
        data[3] = cmd.value
        if args:
            data += args

        if len(data) < 10:
            for i in range(10 - len(data)):
                data.append(0x00)

        resp = funcs[func_name](fcp, cls.__OUI, data)
        if resp[0:4] != data[0:4]:
            raise ValueError('Unexpected response against request.')

        return resp[4:]

    @classmethod
    def command_set(cls, fcp: Hinawa.FwFcp, cmd: VendorCmd, args: bytearray):
        return cls.__command_req(fcp, cmd, args, 'set')

    @classmethod
    def command_get(cls, fcp: Hinawa.FwFcp, cmd: VendorCmd, args: bytearray):
        return cls.__command_req(fcp, cmd, args, 'get')


class MicCmd():
    __TARGETS = {
        'mic-1':    0x00,
        'mic-2':    0x01,
    }

    __PARAMS = {
        True:   0x60,
        False:  0x70,
    }

    @classmethod
    def get_mic_labels(cls):
        return list(cls.__TARGETS.keys())

    @classmethod
    def __command_set(cls, fcp, cmd, target, params, val):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for mic.')
        if val not in params:
            raise ValueError('Invalid ardument for mic command.')
        args = bytearray(3)
        args[0] = 0x80
        args[1] = cls.__TARGETS[target]
        args[2] = params[val]
        ApogeeProtocol.command_set(fcp, cmd, args)

    @classmethod
    def __command_get(cls, fcp, cmd, target, params):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for mic.')
        args = bytearray(3)
        args[0] = 0x80
        args[1] = cls.__TARGETS[target]
        args[2] = 0xff
        resp = ApogeeProtocol.command_get(fcp, cmd, args)
        for key, val in params.items():
            if resp[2] == val:
                return key
        raise OSError('Unexpected response for mic command.')

    @classmethod
    def set_polarity(cls, fcp: Hinawa.FwFcp, target: str, enable: bool):
        cls.__command_set(fcp, VendorCmd.MIC_POLARITY, target,
                          cls.__PARAMS, enable)

    @classmethod
    def get_polarity(cls, fcp: Hinawa.FwFcp, target: str):
        return cls.__command_get(fcp, VendorCmd.MIC_POWER, target,
                                 cls.__PARAMS)

    @classmethod
    def set_power(cls, fcp: Hinawa.FwFcp, target: str, enable: bool):
        cls.__command_set(fcp, VendorCmd.MIC_POLARITY, target,
                          cls.__PARAMS, enable)

    @classmethod
    def get_power(cls, fcp: Hinawa.FwFcp, target: str):
        return cls.__command_get(fcp, VendorCmd.MIC_POWER, target,
                                 cls.__PARAMS)


class InputCmd():
    __ADDR_IN_METERS = 0xfffff0080004

    __TARGETS = {
        'analog-1': 0x00,
        'analog-2': 0x01,
    }

    __ATTRS = {
        '+4dB':     0x60,
        '-10dB':    0x70,
    }

    __CLICKLESS = {
        True:   0x60,
        False:  0x70,
    }

    __LEVELS = {
        'line':         0x60,
        'instrument':   0x70,
    }

    __SRCS = {
        'mic':  0x60,
        'line': 0x70,
    }

    @classmethod
    def get_target_labels(cls):
        return list(cls.__TARGETS.keys())

    @classmethod
    def get_attr_labels(cls):
        return list(cls.__ATTRS.keys())

    @classmethod
    def __command_set(cls, fcp, cmd, target, params, key):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for input.')
        if key not in params:
            raise ValueError('Invalid argument for input command.')
        args = bytearray(3)
        args[0] = 0x80
        args[1] = cls.__TARGETS[target]
        args[2] = params[key]
        ApogeeProtocol.command_set(fcp, cmd, args)

    @classmethod
    def __command_get(cls, fcp, cmd, target, params):
        if target not in cls.__TARGETS:
            raise ValueError('Invalid argument for input.')
        args = bytearray(3)
        args[0] = 0x80
        args[1] = cls.__TARGETS[target]
        args[2] = 0xff
        resp = ApogeeProtocol.command_get(fcp, cmd, args)
        for key, val in params.items():
            if val == resp[2]:
                return key
        raise OSError('Unexpected response for input command.')

    @classmethod
    def set_attr(cls, fcp: Hinawa.FwFcp, target: str, attr: str):
        cls.__command_set(fcp, VendorCmd.IN_ATTR, target, cls.__ATTRS, attr)

    @classmethod
    def get_attr(cls, fcp: Hinawa.FwFcp, target: str):
        return cls.__command_get(fcp, VendorCmd.IN_ATTR, target, cls.__ATTRS)

    @classmethod
    def get_level_labels(cls):
        return list(cls.__LEVELS.keys())

    @classmethod
    def set_level(cls, fcp: Hinawa.FwFcp, target: str, level: str):
        cls.__command_set(fcp, VendorCmd.IN_LEVEL, target, cls.__LEVELS, level)

    @classmethod
    def get_level(cls, fcp: Hinawa.FwFcp, target: str):
        return cls.__command_get(fcp, VendorCmd.IN_LEVEL, target, cls.__LEVELS)

    @classmethod
    def get_src_labels(cls):
        return list(cls.__SRCS.keys())

    @classmethod
    def set_src(cls, fcp: Hinawa.FwFcp, target: str, src: str):
        cls.__command_set(fcp, VendorCmd.IN_SRC, target, cls.__SRCS, src)

    @classmethod
    def get_src(cls, fcp: Hinawa.FwFcp, target: str):
        return cls.__command_get(fcp, VendorCmd.IN_SRC, target, cls.__SRCS)

    @classmethod
    def set_clickless(cls, fcp: Hinawa.FwFcp, enable: bool):
        args = bytearray(3)
        args[0] = 0x80
        args[1] = 0x00
        args[2] = cls.__CLICKLESS[enable]
        ApogeeProtocol.command_set(fcp, VendorCmd.IN_CLICKLESS, args)

    @classmethod
    def get_clickless(cls, fcp: Hinawa.FwFcp):
        args = bytearray(3)
        args[0] = 0x80
        args[1] = 0x00
        args[2] = 0xff
        resp = ApogeeProtocol.command_get(fcp, VendorCmd.IN_CLICKLESS, args)
        for key, val in cls.__CLICKLESS.items():
            if val == resp[2]:
                return key
        raise OSError('Unexpected response for clickless operation.')

    @classmethod
    def set_gain(cls, fcp: Hinawa.FwFcp, target: str, db: float):
        args = bytearray(3)
        args[0] = 0x80
        args[1] = cls.__TARGETS[target]
        args[2] = int(db * 0x4f / 75)
        ApogeeProtocol.command_set(fcp, VendorCmd.IN_GAIN, args)

    @classmethod
    def get_gain(cls, fcp: Hinawa.FwFcp, target: str):
        args = bytearray(3)
        args[0] = 0x80
        args[1] = cls.__TARGETS[target]
        args[2] = 0xff
        # TODO: this doesn't work.
        resp = ApogeeProtocol.command_get(fcp, VendorCmd.IN_GAIN, args)
        return float(resp[2] * 75 / 0x4f)

    @classmethod
    def get_meters(cls, unit: Hinawa.FwNode):
        req = Hinawa.FwReq()
        frames = bytearray(8)
        frames = req.transaction(unit.get_node(),
                                 Hinawa.FwTcode.READ_BLOCK_REQUEST,
                                 cls.__ADDR_IN_METERS, 8, frames)
        vals = unpack('>2I', frames)
        meters = {
            'analog-1': vals[0],
            'alaong-2': vals[1],
        }
        return meters


class OutputCmd():
    __ATTRS = {
        'instrument':   0x60,
        '-10dB':        0x70,
    }

    __MUTES = {
        False:  0x60,
        True:   0x70,
    }

    __SRCS = {
        'stream-1/2':   0x60,
        'mixer-1/2':    0x70,
    }

    @classmethod
    def get_attr_labels(cls):
        return list(cls.__ATTRS.keys())

    @classmethod
    def __command_set(cls, fcp, cmd, params, key):
        if key not in params:
            raise ValueError('Invalid argument for output command.')
        args = bytearray(3)
        args[0] = 0x80
        args[1] = 0x80
        args[2] = params[key]
        ApogeeProtocol.command_set(fcp, cmd, args)

    @classmethod
    def __command_get(cls, fcp, cmd, params):
        args = bytearray(3)
        args[0] = 0x80
        args[1] = 0x80
        args[2] = 0xff
        resp = ApogeeProtocol.command_get(fcp, cmd, args)
        for key, val in params.items():
            if resp[2] == val:
                return key
        raise OSError('Unexpected response to output command.')

    @classmethod
    def set_attr(cls, fcp: Hinawa.FwFcp, attr: str):
        cls.__command_set(fcp, VendorCmd.OUT_ATTR, cls.__ATTRS,
                          attr)

    @classmethod
    def get_attr(cls, fcp: Hinawa.FwFcp):
        return cls.__command_get(fcp, VendorCmd.OUT_ATTR, cls.__ATTRS)

    @classmethod
    def set_mute(cls, fcp: Hinawa.FwFcp, enable: bool):
        cls.__command_set(fcp, VendorCmd.OUT_MUTE, cls.__MUTES, enable)

    @classmethod
    def get_mute(cls, fcp: Hinawa.FwFcp):
        return cls.__command_get(fcp, VendorCmd.OUT_MUTE, cls.__MUTES)

    @classmethod
    def get_out_src_labels(cls):
        return list(cls.__SRCS.keys())

    @classmethod
    def set_out_src(cls, fcp: Hinawa.FwFcp, src: str):
        cls.__command_set(fcp, VendorCmd.OUT_SRC, cls.__SRCS, src)

    @classmethod
    def get_out_src(cls, fcp: Hinawa.FwFcp):
        return cls.__command_get(fcp, VendorCmd.OUT_SRC, cls.__SRCS)

    @classmethod
    def set_volume(cls, fcp: Hinawa.FwFcp, db: float):
        val = int(-db * 0x64 / 48)
        args = bytearray(3)
        args[0] = 0x80
        args[1] = 0x80
        args[2] = val
        ApogeeProtocol.command_set(fcp, VendorCmd.OUT_VOLUME, args)

    @classmethod
    def get_volume(cls, fcp: Hinawa.FwFcp):
        args = bytearray(3)
        args[0] = 0x80
        args[1] = 0x80
        args[2] = 0xff
        resp = ApogeeProtocol.command_get(fcp, VendorCmd.OUT_VOLUME, args)
        return float(resp[2] * -48 / 0x64)


class MixerCmd():
    __ADDR_SRC_LEVELS = 0xfffff0080404

    __SRCS = {
        'stream-1': 0x00,
        'stream-2': 0x01,
        'analog-1': 0x10,
        'analog-2': 0x11,
    }

    @classmethod
    def __build_args(cls, src, ch):
        if src not in cls.__SRCS:
            raise ValueError('Invalid argument for source of mixer.')
        if ch not in (0, 1):
            raise ValueError('Invalid argument for channel of mixer.')
        args = bytearray()
        args.append(cls.__SRCS[src])
        args.append(ch)
        return args

    @classmethod
    def get_mixer_src_labels(cls):
        return list(cls.__SRCS.keys())

    @classmethod
    def set_src_gain(cls, fcp: Hinawa.FwFcp, src: str, ch: int, db: float):
        args = cls.__build_args(src, ch)

        val = int((db + 48) * 0x3fff / 48)
        vals = pack('>H', val)
        args.extend(vals)
        ApogeeProtocol.command_set(fcp, VendorCmd.MIXER_SRC, args)

    @classmethod
    def get_src_gain(cls, fcp: Hinawa.FwFcp, src: str, ch: int):
        args = cls.__build_args(src, ch)
        args.append(0xff)
        args.append(0xff)
        resp = ApogeeProtocol.command_get(fcp, VendorCmd.MIXER_SRC, args)
        val = unpack('>H', resp[2:4])[0]
        return float(48 * val / 0x3fff - 48)

    @classmethod
    def get_meters(cls, unit: Hinawa.FwNode):
        req = Hinawa.FwReq()
        frames = bytearray(16)
        frames = req.transaction(unit.get_node(),
                                 Hinawa.FwTcode.READ_BLOCK_REQUEST,
                                 cls.__ADDR_SRC_LEVELS, 16, frames)
        vals = unpack('>4I', frames)
        meters = {
            'stream-1': vals[0],
            'stream-2': vals[1],
            'analog-1': vals[2],
            'alaong-2': vals[3],
        }
        return meters


class DisplayCmd():
    __TARGETS = {
        'output':   0x60,
        'input':    0x70,
    }

    __OVERHOLDS = {
        'infinite': 0x60,
        '2sec':     0x70,
    }

    __FOLLOWS = {
        False:   0x60,
        True:    0x70,
    }

    @classmethod
    def __command_set(cls, fcp, cmd, params, key):
        if key not in params:
            raise ValueError('Invalid argument for display command.')
        args = bytearray(3)
        args[0] = 0x80
        args[1] = 0x00
        args[2] = params[key]
        ApogeeProtocol.command_set(fcp, cmd, args)

    @classmethod
    def __command_get(cls, fcp, cmd, params):
        args = bytearray(3)
        args[0] = 0x80
        args[1] = 0x00
        args[2] = 0xff
        resp = ApogeeProtocol.command_get(fcp, cmd, args)
        for key, val in params.items():
            if resp[2] == val:
                return key
        raise OSError('Unexpected response for display command.')

    @classmethod
    def get_target_labels(cls):
        return list(cls.__TARGETS.keys())

    @classmethod
    def set_target(cls, fcp: Hinawa.FwFcp, target: str):
        cls.__command_set(fcp, VendorCmd.DISPLAY_TARGET, cls.__TARGETS, target)

    @classmethod
    def get_target(cls, fcp: Hinawa.FwFcp):
        return cls.__command_get(fcp, VendorCmd.DISPLAY_TARGET, cls.__TARGETS)

    @classmethod
    def get_overhold_labels(cls):
        return list(cls.__OVERHOLDS.keys())

    @classmethod
    def set_overhold(cls, fcp: Hinawa.FwFcp, overhold: str):
        cls.__command_set(fcp, VendorCmd.DISPLAY_OVERHOLD, cls.__OVERHOLDS,
                          overhold)

    @classmethod
    def get_overhold(cls, fcp: Hinawa.FwFcp):
        return cls.__command_get(fcp, VendorCmd.DISPLAY_OVERHOLD,
                                 cls.__OVERHOLDS)

    @classmethod
    def set_follow(cls, fcp: Hinawa.FwFcp, enable: bool):
        cls.__command_set(fcp, VendorCmd.DISPLAY_FOLLOW, cls.__FOLLOWS, enable)

    @classmethod
    def get_follow(cls, fcp: Hinawa.FwFcp):
        return cls.__command_get(fcp, VendorCmd.DISPLAY_FOLLOW, cls.__FOLLOWS)

    @classmethod
    def reset_meters(cls, fcp: Hinawa.FwFcp):
        cls.__command_set(fcp, VendorCmd.DISPLAY_CLEAR, {True: 0x70}, True)


class KnobCmd():
    __KNOBS = {
        # label: (selected, val_pos)
        'OUT':  (0x00, 5),
        'IN-1': (0x01, 6),
        'IN-2': (0x02, 7),
    }

    @classmethod
    def get_states(cls, fcp: Hinawa.FwFcp):
        args = bytearray()
        args.append(VendorCmd.HW_STATUS.value)
        resp = ApogeeProtocol.command_get(fcp, VendorCmd.HW_STATUS, None)
        states = {}
        for label, params in cls.__KNOBS.items():
            selected, val_pos = params
            states[label] = resp[val_pos]
            if selected == resp[3]:
                states['knob'] = label
        states['mute'] = bool(resp[2] > 0)
        return states



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.ta1394.general import AvcGeneral

__all__ = ['AvcCcm']


class AvcCcm():
    PLUG_MODE = ('unit', 'subunit')
    PLUG_UNIT_TYPE = ('isoc', 'external')

    @classmethod
    def get_unit_signal_addr(cls, type, plug):
        if type not in cls.PLUG_UNIT_TYPE:
            raise ValueError('Invalid argument for plug unit type')
        if plug >= 30:
            raise ValueError('Invalid argument for plug number')
        addr = bytearray()
        addr.append(0xff)
        if type == 'isoc':
            addr.append(plug)
        else:
            addr.append(0x80 + plug)
        return addr

    @classmethod
    def get_subunit_signal_addr(cls, type, id, plug):
        if type not in AvcGeneral.SUBUNIT_TYPES:
            raise ValueError('Invalid argument for subunit type')
        if plug >= 30:
            raise ValueError('Invalid argument for plug number')
        addr = bytearray()
        addr.append((AvcGeneral.SUBUNIT_TYPES.index(type) << 3) | id)
        addr.append(plug)
        return addr

    @classmethod
    def compare_addrs(cls, a, b):
        if a['mode'] == b['mode'] == 'unit':
            if a['data']['type'] == b['data']['type'] and \
               a['data']['plug'] == b['data']['plug']:
                return True
        elif a['mode'] == b['mode'] == 'subunit':
            if a['data']['type'] == b['data']['type'] and \
               a['data']['id'] == b['data']['id'] and \
               a['data']['plug'] == b['data']['plug']:
                return True
        return False

    @classmethod
    def parse_signal_addr(cls, addr):
        info = {}
        data = {}
        if addr[0] == 0xff:
            info['mode'] = 'unit'
            if addr[1] & 0x80:
                data['type'] = 'external'
                data['plug'] = addr[1] - 0x80
            else:
                data['type'] = 'isoc'
                data['plug'] = addr[1]
        else:
            info['mode'] = 'subunit'
            data['type'] = AvcGeneral.SUBUNIT_TYPES[addr[0] >> 3]
            data['id'] = addr[0] & 0x07
            data['plug'] = addr[1]
        info['data'] = data
        return info

    @classmethod
    def set_signal_source(cls, fcp, src, dst):
        args = bytearray()
        args.append(0x00)
        args.append(0xff)
        args.append(0x1a)
        args.append(0x0f)
        args.append(src[0])
        args.append(src[1])
        args.append(dst[0])
        args.append(dst[1])
        return AvcGeneral.command_control(fcp, args)

    @classmethod
    def get_signal_source(cls, fcp, dst):
        args = bytearray()
        args.append(0x01)
        args.append(0xff)
        args.append(0x1a)
        args.append(0xff)
        args.append(0xff)
        args.append(0xfe)
        args.append(dst[0])
        args.append(dst[1])
        params = AvcGeneral.command_status(fcp, args)
        src = params[4:6]
        return cls.parse_signal_addr(src)

    @classmethod
    def ask_signal_source(cls, fcp, src, dst):
        args = bytearray()
        args.append(0x02)
        args.append(0xff)
        args.append(0x1a)
        args.append(0xff)
        args.append(src[0])
        args.append(src[1])
        args.append(dst[0])
        args.append(dst[1])
        AvcGeneral.command_inquire(fcp, args)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.fireface.ff_status_reg import FFClkLabels

__all__ = ['FFOptionReg']


class FFOptionReg():
    __MULTIPLE_OPTION_MASKS = {
        'line-in': {
            'lo-gain':      (0x00000008, 0x00000000, 0x00000000),
            '-10dB':        (0x00000020, 0x00000002, 0x00000000),
            '+4dB':         (0x00000010, 0x00000003, 0x00000000),
        },
        'line-out': {
            '-10dB':        (0x00001000, 0x00000008, 0x00000000),
            '+4dB':         (0x00000800, 0x00000018, 0x00000000),
            'hi-gain':      (0x00000400, 0x00000010, 0x00000000),
        },
        'primary-clk-src': {
            FFClkLabels.ADAT1.value:    (0x00000000, 0x00000000, 0x00000000),
            FFClkLabels.ADAT2.value:    (0x00000000, 0x00000000, 0x00000400),
            FFClkLabels.SPDIF.value:    (0x00000000, 0x00000000, 0x00000c00),
            FFClkLabels.WDCLK.value:    (0x00000000, 0x00000000, 0x00001400),
        },
    }

    # bit flags for boolean value.
    __SINGLE_OPTION_MASKS = {
        'auto-sync': {
            'internal':     (0x00000000, 0x00000000, 0x00000001),
            'base-441':     (0x00000000, 0x00000000, 0x00000002),
            'base-480':     (0x00000000, 0x00000000, 0x00000004),
            'double':       (0x00000000, 0x00000000, 0x00000008),
            'quadruple':    (0x00000000, 0x00000000, 0x00000010),
        },
        'front-input': {
            'in-1':         (0x00000000, 0x00000800, 0x00000000),
            'in-7':         (0x00000000, 0x00000020, 0x00000000),
            'in-8':         (0x00000000, 0x00000080, 0x00000000),
        },
        'rear-input': {
            'in-1':         (0x00000000, 0x00000004, 0x00000000),
            'in-7':         (0x00000000, 0x00000040, 0x00000000),
            'in-8':         (0x00000000, 0x00000100, 0x00000000),
        },
        'phantom-power': {
            'mic-7':        (0x00000001, 0x00000000, 0x00000000),
            'mic-8':        (0x00000080, 0x00000000, 0x00000000),
            'mic-9':        (0x00000002, 0x00000000, 0x00000000),
            'mic-10':       (0x00000100, 0x00000000, 0x00000000),
        },
        'spdif-in': {
            'optical':      (0x00000000, 0x00000000, 0x00000200),
            'track-maker':  (0x00000000, 0x00000000, 0x40000000),
        },
        'spdif-out': {
            'professional': (0x00000000, 0x00000000, 0x00000020),
            'emphasis':     (0x00000000, 0x00000000, 0x00000040),
            'non-audio':    (0x00000000, 0x00000000, 0x00000080),
            'optical':      (0x00000000, 0x00000000, 0x00000100),
        },
        'instruments': {
            'drive':        (0x00000200, 0x00000200, 0x00000000),
            'limit':        (0x00000000, 0x00000000, 0x00010000),
            'speaker-emu':  (0x00000004, 0x00000000, 0x00000000),
        },
        'wdclk-out': {
            'single-speed': (0x00000000, 0x00000000, 0x00004000),
        },
        'midi-low-addr': {
            '0x00000000':   (0x00000000, 0x00000000, 0x04000000),
            '0x00000080':   (0x00000000, 0x00000000, 0x08000000),
            '0x00000100':   (0x00000000, 0x00000000, 0x10000000),
            '0x00000180':   (0x00000000, 0x00000000, 0x20000000),
        },
        'in-error': {
            'continue':     (0x00000000, 0x00000000, 0x80000000),
        }
    }

    # Helper functions for multiple value options.
    @classmethod
    def get_multiple_option_labels(cls):
        return cls.__MULTIPLE_OPTION_MASKS.keys()

    @classmethod
    def get_multiple_option_value_labels(cls, target):
        if target not in cls.__MULTIPLE_OPTION_MASKS:
            raise ValueError('Invalid argument for multi option.')
        return cls.__MULTIPLE_OPTION_MASKS[target].keys()

    @classmethod
    def build_multiple_option(cls, quads, target, val):
        if target not in cls.__MULTIPLE_OPTION_MASKS:
            raise ValueError('Invalid argument for multi option.')
        if val not in cls.__MULTIPLE_OPTION_MASKS[target]:
            raise ValueError('Invalid argument for value of multi option.')
        elems = cls.__MULTIPLE_OPTION_MASKS[target]
        for name, flags in elems.items():
            for i, flag in enumerate(flags):
                quads[i] &= ~flag
        for i, flag in enumerate(elems[val]):
            quads[i] |= flag

    @classmethod
    def parse_multiple_option(cls, quads, target):
        if target not in cls.__MULTIPLE_OPTION_MASKS:
            raise ValueError('Invalid argument for multi option.')
        elems = cls.__MULTIPLE_OPTION_MASKS[target]
        masks = [0x00] * 3
        for name, flags in elems.items():
            for i, flag in enumerate(flags):
                masks[i] |= flag
        for val, flags in elems.items():
            for i, flag in enumerate(flags):
                if quads[i] & masks[i] != flag:
                    break
            else:
                return val
        raise OSError('Unexpected state of bit flags.')

    # Helper functions to handle boolen options.
    @classmethod
    def get_single_option_labels(cls):
        return cls.__SINGLE_OPTION_MASKS.keys()

    @classmethod
    def get_single_option_item_labels(cls, target):
        if target not in cls.__SINGLE_OPTION_MASKS:
            raise ValueError('Invalid argument for bool option.')
        return cls.__SINGLE_OPTION_MASKS[target].keys()

    @classmethod
    def build_single_option(cls, quads, target, item, enable):
        if target not in cls.__SINGLE_OPTION_MASKS:
            raise ValueError('Invalid argument for bool option.')
        if item not in cls.__SINGLE_OPTION_MASKS[target]:
            raise ValueError('Invalid argument for item of bool option.')
        masks = cls.__SINGLE_OPTION_MASKS[target][item]
        for i, mask in enumerate(masks):
            quads[i] &= ~mask
            if enable:
                quads[i] |= mask

    @classmethod
    def parse_single_option(cls, quads, target, item):
        if target not in cls.__SINGLE_OPTION_MASKS:
            raise ValueError('Invalid argument for bool option.')
        if item not in cls.__SINGLE_OPTION_MASKS[target]:
            raise ValueError('Invalid argument for item of bool option.')
        masks = cls.__SINGLE_OPTION_MASKS[target][item]
        for i, mask in enumerate(masks):
            if quads[i] & mask != mask:
                return False
        return True



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import unpack

from hinawa_utils.ieee1394.config_rom_parser import Ieee1394ConfigRomParser

__all__ = ['Dg00xConfigRomParser']


class Dg00xConfigRomParser(Ieee1394ConfigRomParser):
    _OUI_MICROSOFT = 0x0050f2

    def __init__(self):
        super().__init__()
        self.add_vendor_dep_handle(self._OUI_MICROSOFT,
                                   self.__handle_microsoft_keys)

    def __handle_microsoft_keys(self, key_id, type_name, data):
        if key_id == 0x01 and type_name == 'LEAF':
            width = data[0] >> 4
            character_set = ((data[0] & 0x0f) << 8) | data[1]
            language = unpack('>H', data[2:4])[0]

            if character_set != 0x00 or language != 0x00:
                raise ValueError('Invalid data in descriptor leaf.')

            content = data[4:].decode('US-ASCII') + '\0'
            return content[:content.find('\0')]

        return None

    def parse_rom(self, data):
        entries = super().parse_rom(data)
        return self.__parse_entries(entries['root-directory'])

    def __parse_entries(self, entries):
        # Typical layout.
        FIELDS = (
            ('NODE_CAPABILITIES',   'node-capabilities'),
            ('HARDWARE_VERSION', 'hardware-version'),
            ('VENDOR', 'vendor-id'),
            ('DESCRIPTOR', 'vendor-name'),
            ('UNIT', None),
        )
        info = {}

        for i, field in enumerate(FIELDS):
            entry = entries[i]
            name, alt = field
            if entry[0] != name:
                raise OSError('Invalid format of config ROM.')
            if name == 'UNIT':
                # Check unit.
                entry = entry[1]
                if (entry[0][0] != 'SPECIFIER_ID' or
                    entry[1][0] != 'VERSION' or
                    entry[2][0] != 'MODEL' or
                        entry[3][0] != 'DESCRIPTOR'):
                    raise ValueError('Invalid data in unit directory.')
                info['model-revision'] = entry[0][1]
                info['model-version'] = entry[1][1]
                info['model-id'] = entry[2][1]
                info['model-name'] = entry[3][1]
            else:
                info[alt] = entry[1]

        return info



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import unpack

import gi
gi.require_version('Hinawa', '3.0')
from gi.repository import Hinawa

__all__ = ['TcatProtocolGeneral']


class TcatProtocolGeneral():
    _BASE_ADDR = 0xffffe0000000
    _MAXIMUM_TRX_LENGTH = 512
    RATE_BITS = {
        0x00:   32000,
        0x01:   44100,
        0x02:   48000,
        0x03:   88200,
        0x04:   96000,
        0x05:   176400,
        0x06:   192000,
    }
    _EXT_RATE_BITS = {
        0x07:   48000,
        0x08:   96000,
        0x09:   192000,
    }
    CLOCK_BITS = {
        0x00:   'aes1',
        0x01:   'aes2',
        0x02:   'aes3',
        0x03:   'aes4',
        0x04:   'aes-any',
        0x05:   'adat',
        0x06:   'tdif',
        0x07:   'word-clock',
        0x08:   'arx1',
        0x09:   'arx2',
        0x0a:   'arx3',
        0x0b:   'arx4',
        0x0c:   'internal',
    }

    def __init__(self, unit, req):
        self._unit = unit

        self._general_layout = self._detect_address_space(req)
        self._version = self._parse_dice_version(req)
        self._clock_source_labels = self._parse_clock_source_names(req)
        self._sampling_rates, self._clock_sources = self._parse_clock_caps(req)

    def write_transactions(self, req, offset, data):
        addr = self._BASE_ADDR + offset

        length = len(data)

        while length > 0:
            count = length
            if count > self._MAXIMUM_TRX_LENGTH:
                count = self._MAXIMUM_TRX_LENGTH
            if count == 4:
                tcode = Hinawa.FwTcode.WRITE_QUADLET_REQUEST
            else:
                tcode = Hinawa.FwTcode.WRITE_BLOCK_REQUEST
            req.transaction(self._unit.get_node(), tcode, addr, count,
                            data[0:count])
            data = data[count:]
            length -= count
            addr += count

    def read_transactions(self, req, offset, length):
        data = bytearray()

        addr = self._BASE_ADDR + offset

        while len(data) < length:
            count = length
            if count > self._MAXIMUM_TRX_LENGTH:
                count = self._MAXIMUM_TRX_LENGTH
            if count == 4:
                tcode = Hinawa.FwTcode.READ_QUADLET_REQUEST
            else:
                tcode = Hinawa.FwTcode.READ_BLOCK_REQUEST
            frames = bytearray(count)
            frames = req.transaction(self._unit.get_node(), tcode, addr, count,
                                     frames)
            data.extend(frames)
            length -= count
            addr += count

        return data

    def _read_section_offset(self, req, section, offset, length):
        offset += self._general_layout[section]['offset']
        return self.read_transactions(req, offset, length)

    def _write_section_offset(self, req, section, offset, data):
        offset += self._general_layout[section]['offset']
        self.write_transactions(req, offset, data)

    def _detect_address_space(self, req):
        PARAMS = (
            ('global',  40, 0x60),
            ('tx',      40, 0x18),
            ('rx',      40, 0x18),
            ('external', 0, 0),
            ('reserved', 0, 0),
        )
        layout = {}

        data = self.read_transactions(req, 0, len(PARAMS) * 8)
        for i, param in enumerate(PARAMS):
            offset = unpack('>I', data[0:4])[0] * 4
            count = unpack('>I', data[4:8])[0] * 4
            data = data[8:]
            if offset >= param[1] or count >= param[2]:
                layout[param[0]] = {
                    'offset':   offset,
                    'length':   count,
                }
        return layout

    def _parse_string_bytes(self, data):
        letters = bytearray()

        for i in range(0, len(data), 4):
            letters.extend(list(reversed(data[0:4])))
            data = data[4:]

        return letters.decode('utf-8').rstrip('\0')

    # GLOBAL_OWNER: global:0x00
    def read_owner_addr(self, req):
        data = self._read_section_offset(req, 'global', 0x00, 8)
        return (unpack('>I', data[0:4])[0] << 32) | unpack('>I', data[4:8])[0]

    # GLOBAL_NOTIFICATION: global:0x08
    def read_latest_notification(self, req):
        data = self._read_section_offset(req, 'global', 0x08, 4)
        return unpack('>I', data)[0]

    # GLOBAL_NICKNAME: global:0x0c
    def write_nickname(self, req, name):
        letters = bytearray(name.encode('utf-8'))
        if len(letters) > 63:
            raise ValueError('The length of name should be within 63 bytes.')
        for i in range(64 - len(letters)):
            letters.append(0x00)

        data = bytearray()
        for i in range(0, len(letters), 4):
            data.extend(list(reversed(letters[0:4])))
            letters = letters[4:]

        self._write_section_offset(req, 'global', 0x0c, data)

    def read_nickname(self, req):
        data = self._read_section_offset(req, 'global', 0x0c, 64)
        return self._parse_string_bytes(data).rstrip()

    def _clock_select_transaction(self, data):
        quads = unpack('>I', data)
        offset = self._general_layout['global']['offset'] + 0x4c
        self._unit.transaction(self._BASE_ADDR + offset, quads, 0x00000020)

    # GLOBAL_CLOCK_SELECT: global:004c
    def get_supported_clock_sources(self):
        return self._clock_sources

    def write_clock_source(self, req, source):
        if source not in self._clock_sources:
            raise ValueError('Invalid argument for clock source.')
        val = {v: k for k, v in self.CLOCK_BITS.items()}[source]

        data = self._read_section_offset(req, 'global', 0x4c, 4)
        if data[3] != val:
            data[3] = val
            self._clock_select_transaction(data)

    def read_clock_source(self, req):
        data = self._read_section_offset(req, 'global', 0x4c, 4)
        val = data[3]
        if (val not in self.CLOCK_BITS or
                self._clock_source_labels[val] == 'Unused'):
            raise OSError('Unexpected return value for clock source.')

        return self.CLOCK_BITS[val]

    def get_supported_sampling_rates(self):
        return self._sampling_rates

    def write_sampling_rate(self, req, rate):
        for index, val in self.RATE_BITS.items():
            if rate == val:
                break
        else:
            for index, val in self._EXT_RATE_BITS.items():
                if rate < val:
                    break
            else:
                raise ValueError('Invalid argument for sampling rate.')

        data = self._read_section_offset(req, 'global', 0x4c, 4)
        if data[2] != index:
            data[2] = index
            self._clock_select_transaction(data)

    def read_sampling_rate(self, req):
        data = self._read_section_offset(req, 'global', 0x4c, 4)
        index = data[2]
        if index in self.RATE_BITS:
            return self.RATE_BITS[index]
        if index in self._EXT_RATE_BITS:
            return self._EXT_RATE_BITS[index]

        raise OSError('Unexpected return value for sampling rate.')

    # GLOBAL_ENABLE: global:0x50
    def read_enabled(self, req):
        data = self._read_section_offset(req, 'global', 0x50, 4)
        return bool(unpack('>I', data))

    # GLOBAL_STATUS: global:0x54
    def read_clock_status(self, req):
        status = {}

        data = self._read_section_offset(req, 'global', 0x54, 4)
        status['locked'] = bool(data[3])
        status['rate'] = self.RATE_BITS[data[2]]

        return status

    # GLOBAL_EXTENDED_STATUS: global 0x58
    def read_external_clock_states(self, req):
        CLOCK_BITS = {
            0x0001: 'aes1',
            0x0002: 'aes2',
            0x0004: 'aes3',
            0x0008: 'aes4',
            0x0010: 'adat',
            0x0020: 'tdif',
            0x0040: 'arx1',
            0x0080: 'arx2',
            0x0100: 'arx3',
            0x0200: 'arx4',
            0x0400: 'word-clock',
        }
        status = {
            'locked':   [],
            'slipped':  [],
        }

        data = self._read_section_offset(req, 'global', 0x58, 4)

        slipped_mask = unpack('>H', data[0:2])[0]
        locked_mask = unpack('>H', data[2:4])[0]
        for bit, clk in CLOCK_BITS.items():
            if bit & slipped_mask:
                status['slipped'].append(clk)
            if bit & locked_mask:
                status['locked'].append(clk)
        return status

    # GLOBAL_SAMPLE_RATE: global:0x5c
    def read_measured_sampling_rate(self, req):
        data = self._read_section_offset(req, 'global', 0x5c, 4)
        return unpack('>I', data)[0]

    # GLOBAL_VERSION: global:0x60
    def _parse_dice_version(self, req):
        if self._general_layout['global']['length'] < 0x64:
            return '1.0.2'
        data = self._read_section_offset(req, 'global', 0x60, 4)
        return '{0}.{1}.{2}.{3}'.format(data[0], data[1], data[2], data[3])

    def get_dice_version(self):
        return self._version

    # GLOBAL_CLOCK_CAPABILITIES: global:0x64
    def _parse_clock_caps(self, req):
        rates = []
        clks = []

        if self._version == '1.0.2':
            return [44100, 48000], ['arx1', 'internal']

        data = self._read_section_offset(req, 'global', 0x64, 4)

        flags = unpack('>H', data[0:2])[0]
        for index, name in self.CLOCK_BITS.items():
            label = self._clock_source_labels[index]
            if (1 << index) & flags and label != 'Unused':
                clks.append(name)

        flags = unpack('>H', data[2:4])[0]
        for index, rate in self.RATE_BITS.items():
            if (1 << index) & flags:
                rates.append(rate)

        return rates, clks

    # GLOBAL_CLOCK_SOURCE_NAMES: global:0x68
    def _parse_clock_source_names(self, req):
        if self._version == '1.0.2':
            names = []
            for key, name in self.CLOCK_BITS.items():
                if name == 'arx1':
                    names.append('Firewire')
                elif name == 'internal':
                    names.append('Internal')
                else:
                    names.append('Unused')
            return names
        data = self._read_section_offset(req, 'global', 0x68, 256)
        return self._parse_string_bytes(data).split('\\')[0:-2]

    def get_clock_source_names(self):
        return self._clock_source_labels

    # TX stream settings.
    def read_tx_params(self, req):
        params = []
        data = self._read_section_offset(req, 'tx', 0x00, 8)
        count = unpack('>I', data[0:4])[0]
        length = unpack('>I', data[4:8])[0] * 4
        for i in range(count):
            offset = 0x08 + length * i
            data = self._read_section_offset(req, 'tx', offset, length)
            pcm_count = unpack('>I', data[4:8])[0]
            pcm_formation = self._parse_string_bytes(data[16:272])
            pcm_formation = pcm_formation.split('\\')[0:pcm_count]
            stream = {
                'iso-channel': unpack('>I', data[0:4])[0],
                'pcm':         pcm_count,
                'midi':        unpack('>I', data[8:12])[0],
                'speed':       unpack('>I', data[12:16])[0],
                'formation':   pcm_formation,
            }
            if length >= 280:
                stream['iec60958'] = {
                    'caps':    unpack('>I', data[272:276])[0],
                    'enable':  unpack('>I', data[276:280])[0],
                }
            params.append(stream)
        return params

    # RX stream settings.
    def read_rx_params(self, req):
        params = []
        data = self._read_section_offset(req, 'rx', 0x00, 8)
        count = unpack('>I', data[0:4])[0]
        length = unpack('>I', data[4:8])[0] * 4
        for i in range(count):
            offset = 0x08 + length * i
            data = self._read_section_offset(req, 'rx', offset, length)
            pcm_count = unpack('>I', data[8:12])[0]
            pcm_formation = self._parse_string_bytes(data[16:272])
            pcm_formation = pcm_formation.split('\\')[0:pcm_count]
            stream = {
                'iso-channel':  unpack('>I', data[0:4])[0],
                'start':        unpack('>I', data[4:8])[0],
                'pcm':          pcm_count,
                'midi':         unpack('>I', data[12:16])[0],
                'formation':    pcm_formation,
            }
            if length >= 280:
                stream['iec60958'] = {
                    'caps':    unpack('>I', data[272:276])[0],
                    'enable':  unpack('>I', data[276:280])[0],
                }
            params.append(stream)
        return params

    # External synchronization status.
    def read_external_sync_clock_source(self, req):
        if ('external' not in self._general_layout or
                self._general_layout['external']['length'] == 0):
            return ''
        data = self._read_section_offset(req, 'external', 0x00, 4)
        val = unpack('>I', data)[0]
        if (val not in self.CLOCK_BITS or
                self._clock_source_labels[val] == 'Unused'):
            return ''
        return self.CLOCK_BITS[val]

    def read_external_sync_locked(self, req):
        if ('external' not in self._general_layout or
                self._general_layout['external']['length'] == 0):
            return ''
        data = self._read_section_offset(req, 'external', 0x04, 4)
        val = unpack('>I', data)[0]
        return bool(val)

    def read_external_sync_rate(self, req):
        if ('external' not in self._general_layout or
                self._general_layout['external']['length'] == 0):
            return ''
        data = self._read_section_offset(req, 'external', 0x08, 4)
        val = unpack('>I', data)[0]
        if val not in self.RATE_BITS:
            raise OSError('Unexpected return value for sampling rate.')
        return self.RATE_BITS[val]

    def read_external_sync_adat_status(self, req):
        if ('external' not in self._general_layout or
                self._general_layout['external']['length'] == 0):
            return ''
        data = self._read_section_offset(req, 'external', 0x0c, 4)
        if not data[3] & 0x10:
            return 0
        return data[3] & 0xf



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2019 Takashi Sakamoto


class FFOutRegs():
    @classmethod
    def get_out_labels(cls, spec: dict):
        labels = []
        for i in range(spec['analog']):
            labels.append('analog-{0}'.format(i + 1))
        for i in range(spec['spdif']):
            labels.append('spdif-{0}'.format(i + 1))
        for i in range(spec['adat']):
            labels.append('adat-{0}'.format(i + 1))
        return labels

    @classmethod
    def calculate_out_offset(cls, spec: dict, target):
        targets = cls.get_out_labels(spec)
        return targets.index(target) * 4



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.motu.motu_protocol_abstract import MotuProtocolAbstract

__all__ = ['MotuProtocolV2']


class MotuProtocolV2(MotuProtocolAbstract):
    # direction/index/mask/shift
    OPT_IFACE_MODE_ATTRS = {
        'in':   (2, 0x03, 0),
        'out':  (2, 0x0c, 2),
    }

    PACKET_SIZE_ATTRS = {
        'in':   (3, 0x80, 7),
        'out':  (3, 0x40, 6),
    }

    def get_supported_sampling_rates(self):
        if self._unit.name in ('828mk2', ):
            return self.SUPPORTED_SAMPLING_RATES_X2
        if self._unit.name in ('Traveler', ):
            return self.SUPPORTED_SAMPLING_RATES_X4

    def get_sampling_rate(self):
        frames = self.read(0x0b14, 4)
        val = (frames[3] & 0x38) >> 3
        return self.SUPPORTED_SAMPLING_RATES_X4[val]

    def set_sampling_rate(self, rate):
        frames = self.read(0x0b14, 4)
        frames[3] &= ~0x00000038
        frames[3] |= self.SUPPORTED_SAMPLING_RATES_X4.index(rate) << 3
        self.write(0x0b14, frames)

    def get_supported_clock_sources(self):
        sources = [
            self.CLOCK_INTERNAL,
            self.CLOCK_WORD_ON_BNC,
            self.CLOCK_SPDIF_ON_COAX,
            self.CLOCK_ADAT_ON_DSUB,
        ]
        mode = self.get_opt_iface_mode('in', 'A')
        if mode == 'S/PDIF':
            sources.append(self.CLOCK_SPDIF_ON_OPT)
        else:
            sources.append(self.CLOCK_ADAT_ON_OPT)

        return sources

    def get_clock_source(self):
        frames = self.read(0x0b14, 4)

        val = frames[3] & 0x07
        if val == 0:
            return self.CLOCK_INTERNAL
        elif val == 1:
            return self.CLOCK_ADAT_ON_OPT
        elif val == 4:
            return self.CLOCK_WORD_ON_BNC
        elif val == 5:
            return self.CLOCK_ADAT_ON_DSUB
        elif val == 2:
            mode = self.get_opt_iface_mode('in', 'A')
            if mode == 'S/PDIF':
                return self.CLOCK_ADAT_ON_OPT
            else:
                return self.CLOCK_SPDIF_ON_COAX

    def set_clock_source(self, source):
        frames = self.read(0x0b14, 4)

        frames[3] &= ~0x07
        if source == self.CLOCK_ADAT_ON_OPT:
            frames[3] |= 0x01
        elif source == self.CLOCK_WORD_ON_BNC:
            frames[3] |= 0x04
        elif source == self.CLOCK_ADAT_ON_DSUB:
            frames[3] |= 0x05
        elif source in (self.CLOCK_SPDIF_ON_OPT, self.CLOCK_SPDIF_ON_COAX):
            mode = self.get_opt_iface_mode('in', 'A')
            if source == self.CLOCK_SPDIF_ON_OPT and mode != 'S/PDIF':
                raise ValueError('This mode is currently unavailable.')
            frames[3] |= 0x02

        self.write(0x0b14, frames)

    def get_supported_opt_iface_indexes(self):
        if self._unit.name == '828mk2':
            return self.SUPPORTED_OPT_IFACE_INDEXES[0:1]
        else:
            return self.SUPPORTED_OPT_IFACE_INDEXES

    def get_opt_iface_mode(self, direction, index):
        frames = self.read(0x0c04, 4)

        pos, mask, shift = self.OPT_IFACE_MODE_ATTRS[direction]
        val = (frames[pos] & mask) >> shift
        if val >= len(self.SUPPORTED_OPT_IFACE_MODES):
            raise OSError('Unexpected value for opt iface mode.')

        if val == 1:
            val = 2
        elif val == 2:
            val = 1
        return self.SUPPORTED_OPT_IFACE_MODES[val]

    def set_opt_iface_mode(self, direction, index, mode):
        frames = self.read(0x0c04, 4)

        idx = self.SUPPORTED_OPT_IFACE_MODES.index(mode)
        if idx == 1:
            idx = 2
        elif idx == 2:
            idx = 1

        pos, mask, shift = self.OPT_IFACE_MODE_ATTRS[direction]
        frames[pos] &= ~mask
        frames[pos] |= idx << shift

        self.write(0x0c04, frames)

        pos, mask, shift = self.PACKET_SIZE_ATTRS[direction]
        frames = self.read(0x0b10, 4)
        frames[pos] &= ~mask
        if mode != 'ADAT':
            frames[pos] |= 1 << shift

        self.write(0x0b10, frames)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from threading import Thread
from struct import pack, unpack
from math import log10, pow

import gi
gi.require_version('GLib', '2.0')
gi.require_version('Hinawa', '3.0')
from gi.repository import GLib, Hinawa

from hinawa_utils.tscm.config_rom_parser import TscmConfigRomParser

__all__ = ['TscmUnit']


class TscmUnit(Hinawa.SndUnit):
    _BASE_ADDR = 0xffff00000000

    supported_sampling_rates = (44100, 48000, 88200, 96000)
    supported_clock_sources = ('Internal', 'Word-clock', 'S/PDIF', 'ADAT')
    supported_coax_sources = ('S/PDIF-1/2', 'Analog-1/2')
    supported_led_status = ('off', 'on')

    __SPECS = {
        'FW-1082': {
            'stream-in-ana-chs':    2,
            'has-opt-iface':        False,
        },
        'FW-1804': {
            'stream-in-ana-chs':    2,
            'has-opt-iface':        True,
        },
        'FW-1884': {
            'stream-in-ana-chs':    8,
            'has-opt-iface':        True,
        },
    }

    __MAX_THRESHOLD = 0x7fff

    def __init__(self, path):
        super().__init__()
        self.open(path)
        if self.get_property('type') != 6:
            raise ValueError('The character device is not for Tascam unit')

        ctx = GLib.MainContext.new()
        self.create_source().attach(ctx)
        self.__unit_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__unit_th = Thread(target=lambda d: d.run(), args=(self.__unit_dispatcher, ))
        self.__unit_th.start()

        node = self.get_node()
        ctx = GLib.MainContext.new()
        node.create_source().attach(ctx)
        self.__node_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__node_th = Thread(target=lambda d: d.run(), args=(self.__node_dispatcher, ))
        self.__node_th.start()

        parser = TscmConfigRomParser()
        info = parser.parse_rom(self.get_node().get_config_rom())
        self.model_name = info['model-name']
        self.__specs = self.__SPECS[self.model_name]

    def release(self):
        self.__unit_dispatcher.quit()
        self.__node_dispatcher.quit()
        self.__unit_th.join()
        self.__node_th.join()

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.release()

    def read_quadlet(self, offset):
        req = Hinawa.FwReq()
        frames = bytearray(4)
        return req.transaction(self.get_node(),
                        Hinawa.FwTcode.READ_QUADLET_REQUEST,
                        self._BASE_ADDR + offset, 4, frames)

    def write_quadlet(self, offset, frames):
        req = Hinawa.FwReq()
        return req.transaction(self.get_node(),
                        Hinawa.FwTcode.WRITE_QUADLET_REQUEST,
                        self._BASE_ADDR + offset, 4, frames)

    def get_firmware_versions(self):
        info = {}
        frames = self.read_quadlet(0x0000)
        info['Register'] = unpack('>I', frames)[0]
        frames = self.read_quadlet(0x0004)
        info['FPGA'] = unpack('>I', frames)[0]
        frames = self.read_quadlet(0x0008)
        info['ARM'] = unpack('>I', frames)[0]
        frames = self.read_quadlet(0x000c)
        info['HW'] = unpack('>I', frames)[0]
        return info

    def set_clock_source(self, src):
        if src not in self.supported_clock_sources:
            raise ValueError(
                'Invalid argument for clock source: {0}'.format(src))
        src = self.supported_clock_sources.index(src) + 1
        frames = self.read_quadlet(0x0228)
        frames = bytearray(frames)
        frames[0] = 0
        frames[1] = 0
        frames[3] = src
        self.write_quadlet(0x0228, frames)

    def get_clock_source(self):
        frames = self.read_quadlet(0x0228)
        index = frames[3] - 1
        if index < 0 or index >= len(self.supported_clock_sources):
            raise OSError('Unexpected value for clock source.')
        return self.supported_clock_sources[index]

    def set_sampling_rate(self, rate):
        if rate not in self.supported_sampling_rates:
            raise ValueError('Invalid argument for sampling rate.')
        if rate == 44100:
            flag = 0x01
        elif rate == 48000:
            flag = 0x02
        elif rate == 88200:
            flag = 0x81
        elif rate == 96000:
            flag = 0x82
        frames = self.read_quadlet(0x0228)
        frames = bytearray(frames)
        frames[3] = flag
        self.write_quadlet(0x0228, frames)

    def get_sampling_rate(self):
        frames = self.read_quadlet(0x0228)
        if (frames[1] & 0x0f) == 0x01:
            rate = 44100
        elif (frames[1] & 0x0f) == 0x02:
            rate = 48000
        else:
            raise OSError('Unexpected value for sampling rate.')
        if (frames[1] & 0xf0) == 0x80:
            rate *= 2
        elif (frames[1] & 0xf0) != 0x00:
            raise OSError('Unexpected value for sampling rate.')
        return rate

    def __set_flag(self, flag, mask):
        data = self.read_quadlet(0x022c)
        flags = data[3]
        data = bytearray(4)
        data[1] = (flags & ~flag) & mask    # removed bits.
        data[2] = (~flags & flag) & mask    # added bits.
        self.write_quadlet(0x022c, data)

    def get_stream_spdif_in_src_labels(self):
        labels = []
        labels.append('coax-iface')
        labels.append('opt-iface')
        return labels

    def set_stream_spdif_in_src(self, src):
        labels = self.get_stream_spdif_in_src_labels()
        if src not in labels:
            raise ValueError('Invalid argumen for source of stream spdif in.')
        flag = labels.index(src)
        self.__set_flag(flag, 0x01)

    def get_stream_spdif_in_src(self):
        labels = self.get_stream_spdif_in_src_labels()
        data = self.read_quadlet(0x022c)
        index = data[3] & 0x01
        return labels[index]

    def get_coax_out_src_labels(self):
        labels = []
        ana_ch = self.__specs['stream-in-ana-chs']
        labels.append('stream-in-{0}/{1}'.format(ana_ch + 1, ana_ch + 2))
        labels.append('analog-out-1/2')
        return labels

    def set_coax_out_src(self, src):
        labels = self.get_coax_out_src_labels()
        if src not in labels:
            raise ValueError('Invalid argument for source of coax out iface.')
        flag = labels.index(src) << 1
        self.__set_flag(flag, 0x02)

    def get_coax_out_src(self):
        labels = self.get_coax_out_src_labels()
        data = self.read_quadlet(0x022c)
        index = (data[3] & 0x02) >> 1
        return labels[index]

    def get_opt_out_src_labels(self):
        labels = []
        if self.__specs['has-opt-iface']:
            ana_ch = self.__specs['stream-in-ana-chs']

            label = 'stream-in-{0}:{1}'.format(ana_ch + 1, ana_ch + 9)
            labels.append(label)

            label = 'stream-in-{0}/{1}'.format(ana_ch + 10, ana_ch + 11)
            labels.append(label)

            labels.append('analog-in-1:8')
            if ana_ch > 2:
                labels.append('analog-out-1:8')

        return labels

    def set_opt_out_src(self, src):
        labels = self.get_opt_out_src_labels()
        if src not in labels:
            print('Invalid argument for source of opt out iface.')
        flag = labels.index(src) << 2
        if flag == 0x0c:
            flag = 0x88
        self.__set_flag(flag, 0x8c)

    def get_opt_out_src(self):
        labels = self.get_opt_out_src_labels()
        data = self.read_quadlet(0x022c)
        index = (data[3] & 0x0c) >> 2
        if data[3] & 0x80:
            index += 1
        print(index)
        return labels[index]

    def set_input_threshold(self, level):
        data = self.read_quadlet(0x0230)
        data = bytearray(data)
        if level == float('-inf'):
            val = 0x0000
        else:
            val = int(self.__MAX_THRESHOLD * pow(10, level / 20))
        chunks = pack('>H', val)
        data[0] = chunks[0]
        data[1] = chunks[1]
        self.write_quadlet(0x0230, data)

    def get_input_threshold(self):
        data = self.read_quadlet(0x0230)
        val = unpack('>H', data[0:2])[0]
        if val == 0x0000:
            return float('-inf')
        return 20 * log10(val / self.__MAX_THRESHOLD)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

import sys
import string
from pathlib import Path
from signal import SIGINT

import gi
gi.require_version('GLib', '2.0')
gi.require_version('Hinawa', '3.0')
from gi.repository import GLib, Hinawa

__all__ = ['CliKit']


class CliKit():
    @staticmethod
    def _seek_snd_unit_from_guid(guid):
        for fullpath in Path('/dev/snd').glob('hw*'):
            fullpath = str(fullpath)
            try:
                unit = Hinawa.SndUnit()
                unit.open(fullpath)
                if unit.get_property('guid') == guid:
                    return fullpath
            except Exception as e:
                pass
            finally:
                del unit
        return None

    @staticmethod
    def _check_hexadecimal(literal):
        if literal.find('0x') == 0:
            literal = literal[2:]
        if len(literal) != 16:
            return False
        for character in literal:
            if character not in string.hexdigits:
                return False
        else:
            return True

    @staticmethod
    def _dump_help(cmdline):
        print('{0} CARD|GUID [FILE|CMD [ARGS]]'.format(cmdline))
        print('  CARD:  the number as ALSA sound card, see /proc/asound/cards.')
        print('  GUID:  global unique ID for your unit.')
        print('  FILE:  path for a file with command list')
        print('  CMD:   issue which you need')
        print('  ARGS:  arguments for the command')

    @staticmethod
    def _dump_commands(cmds):
        print('Available commands:')
        for name in cmds.keys():
            print('  {0}'.format(name))

    @classmethod
    def seek_snd_unit_path(cls):
        args = sys.argv
        if len(args) > 1:
            identity = args[1]
            # Assume as sound card number if it's digit literal.
            if identity.isdigit():
                return '/dev/snd/hwC{0}D0'.format(identity)
            # Assume as GUID on IEEE 1394 bus if it's hexadecimal literal.
            elif cls._check_hexadecimal(identity):
                return cls._seek_snd_unit_from_guid(int(identity, base=16))
        cls._dump_help(args[0])
        return None

    @classmethod
    def handle_unix_signal(cls, unit):
        del unit

    @classmethod
    def dispatch_command(cls, unit, cmds):
        args = sys.argv
        if len(args) > 2:
            # Install signal handler to cancel event dispatcher.
            GLib.unix_signal_add(GLib.PRIORITY_HIGH, SIGINT,
                                 cls.handle_unix_signal, unit)

            if args[2] in cmds:
                cmd = args[2]
                return cmds[cmd](unit, args[3:])
            path = Path(args[2])
            if path.is_file():
                with path.open(mode='r') as fh:
                    for i, line in enumerate(fh):
                        args = line.rstrip().split(' ')
                        if len(args) == 0:
                            continue

                        cmd = args[0]
                        if len(cmd) == 0:
                            continue
                        if cmd[0] == '#':
                            continue
                        if cmd not in cmds:
                            print('Invalid command in {0}: {1}: {2}'.format(
                                str(path), i, cmd))
                            return False

                        if not cmds[cmd](unit, args[1:]):
                            print('Invalid arguments in {0}:{1}: {2}'.format(
                                str(path), i, cmd))
                            return False
                return True
        cls._dump_commands(cmds)
        return False
